/*****************************************************************************
 * audio.c: audio decoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008-2010 M2X BV
 * $Id: 5f5e39c1c322f2dde0675303d4526ec65a32fcd7 $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *          Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_aout.h>
#include <vlc_codec.h>

#include "davinci.h"

#include <ti/sdo/ce/audio/auddec.h>

#include <assert.h>

/* Define this to use PTS from the block corresponding to the first byte in
 * the input buffer instead of the PTS from the block corresponding to the
 * current p_block (so bytes at the end of the buffer or right after) */
//#define TRACK_PTS

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static aout_buffer_t *DecodeAudioBlock( decoder_t *, block_t ** );

/*****************************************************************************
 * Audio decoder
 *****************************************************************************/
#ifdef TRACK_PTS
struct block_tracker_t
{
    mtime_t i_pts;
    ssize_t i_left;
};
#endif

struct decoder_sys_t
{
    char *psz_ti_engine;
    Engine_Handle e;
    AUDDEC_Handle d;

    XDM_BufDesc in;
    XDM_BufDesc out;

    audio_date_t date;
    int i_buffer;

#ifdef TRACK_PTS
    struct block_tracker_t p_tracked_blocks[100];
    int i_tracked_blocks;
#endif
};

/*****************************************************************************
 *
 *****************************************************************************/
int OpenAudioDecoder( vlc_object_t *p_this )
{
    decoder_t *p_dec = (decoder_t *)p_this;
    decoder_sys_t *p_sys;
    Engine_Error err;
    AUDDEC_Params params;
    int i_cat, i_xdm;
    const char *psz_codec;
    const char *psz_namecodec;

    if( !GetDavinciDecoder( p_dec->fmt_in.i_codec, &i_cat, &i_xdm,
                            &psz_codec, &psz_namecodec ) )
    {
        msg_Dbg( p_dec, "Unsupported audio codec : %4.4s",(char*)&p_dec->fmt_in.i_codec );
        return VLC_EGENERIC;
    }

    if( i_cat != AUDIO_ES ) return VLC_EGENERIC;

    /* Allocate our private structure */
    p_dec->p_sys = p_sys = (decoder_sys_t *)calloc( 1, sizeof( decoder_sys_t ) );
    if( !p_dec->p_sys )
        return VLC_ENOMEM;

    p_sys->psz_ti_engine = var_CreateGetString( p_dec, DEC_CFG_PREFIX "engine" );
    if( !p_sys->psz_ti_engine )
    {
        free( p_sys );
        return VLC_ENOMEM;
    }

    /* Initialize the codec engine */
    davinci_RuntimeInit( p_this );

    /* Create an engine handle */
    p_sys->e = Engine_open( p_sys->psz_ti_engine, NULL /*&Engine_ATTRS*/, &err );
    if( err != Engine_EOK )
    {
        msg_Err( p_dec, "Error while opening engine `%s': %s",
                 p_sys->psz_ti_engine, ppsz_engine_error[err] );
        goto error;
    }

    davinci_PrintAvailableAlgorithms( p_this, p_sys->psz_ti_engine );

    /* Create audio decoder */
    params.size = sizeof( params );
    params.maxSampleRate = 0; /* Use default. in Hz */
    params.maxBitrate = 0; /* Use default. in kbps */
    params.maxNoOfCh = 0; /* Use default */
    params.dataEndianness = XDM_BYTE; /* FIXME: ? */

    p_sys->d = AUDDEC_create( p_sys->e, (String)psz_codec, &params );
    if( !p_sys->d )
    {
        msg_Err( p_dec, "Failed to create audio decoder (%s) for %s",
                 psz_codec, psz_namecodec );
        goto error;
    }

    /* Set output properties */

    /* Set callbacks */
    p_dec->pf_decode_audio = DecodeAudioBlock;

    if( p_dec->fmt_in.audio.i_rate )
    {
        aout_DateInit( &p_sys->date, p_dec->fmt_in.audio.i_rate );
        aout_DateSet( &p_sys->date, 0 );
    }
    p_dec->fmt_out.i_cat = AUDIO_ES;
    p_dec->fmt_out.i_codec =
        p_dec->fmt_out.audio.i_format = AOUT_FMT_S16_NE;
    p_dec->fmt_out.audio.i_bitspersample = 16;

    return VLC_SUCCESS;

error:
    if( p_sys->e ) Engine_close( p_sys->e );
    free( p_sys->psz_ti_engine );
    free( p_sys );
    return VLC_EGENERIC;
}

/*****************************************************************************
 *
 *****************************************************************************/
void CloseAudioDecoder( vlc_object_t *p_this )
{
    decoder_t *p_dec = (decoder_t *)p_this;
    decoder_sys_t *p_sys = p_dec->p_sys;

    /* Close our codec handle */
    AUDDEC_delete( p_sys->d );

    /* Close our engine handle */
    Engine_close( p_sys->e );

    /* Exit the codec engine */
    davinci_RuntimeExit( p_this );

    /* Free 'DaVinci compliant' buffers */
    davinci_FreeBuffer( &p_sys->in );
    davinci_FreeBuffer( &p_sys->out );

    free( p_sys->psz_ti_engine );
    free( p_sys );
}

static int InitXDMAudioDecoder( decoder_t *p_dec, AUDDEC_DynamicParams *dparams )
{
    decoder_sys_t *p_sys = p_dec->p_sys;
    AUDDEC_Status status;

    /* Configure audio decoder */
    /*if( AUDDEC_control( p_sys->d, XDM_RESET, &dparams, &status )
        != AUDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to reset decoder" );
        goto error;
    }*/

    dparams->outputFormat = IAUDIO_INTERLEAVED;
    if( AUDDEC_control( p_sys->d, XDM_SETPARAMS, dparams, &status ) != AUDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to set dynamic decoder parameters" );
        return VLC_EGENERIC;
    }

    if( AUDDEC_control( p_sys->d, XDM_GETBUFINFO, dparams, &status ) != AUDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to get buffer info" );
        return VLC_EGENERIC;
    }

    /* Allocate input buffer */
    if( davinci_AllocateBuffer( status.bufInfo.minNumInBufs,
                                status.bufInfo.minInBufSize, &p_sys->in ) != VLC_SUCCESS )
    {
        msg_Err( p_dec, "Failed to allocate input buffer" );
        return VLC_EGENERIC;
    }

    /* Allocate output buffer */
    if( davinci_AllocateBuffer( status.bufInfo.minNumOutBufs,
                                status.bufInfo.minOutBufSize, &p_sys->out ) != VLC_SUCCESS )
    {
        msg_Err( p_dec, "Failed to allocate output buffer" );
        return VLC_EGENERIC;
    }

    return VLC_EGENERIC;
}

/*****************************************************************************
 *
 *****************************************************************************/
static aout_buffer_t *DecodeAudioBlock( decoder_t *p_dec, block_t **pp_block )
{
    decoder_sys_t *p_sys = p_dec->p_sys;
    block_t *p_block;
    aout_buffer_t *p_out = NULL;
    AUDDEC_InArgs in_args;
    AUDDEC_OutArgs out_args;
    AUDDEC_DynamicParams dparams;
    AUDDEC_Status status;
    int i;
    bool b_reject = false;

    if( !pp_block || !*pp_block ) return NULL;
    p_block = *pp_block;

    if( p_block->i_flags & (BLOCK_FLAG_DISCONTINUITY|BLOCK_FLAG_CORRUPTED) )
    {
        block_Release( p_block );
        p_sys->i_buffer = 0;
        return NULL;
    }

    memset( &in_args, 0, sizeof( in_args ) );
    memset( &out_args, 0, sizeof( out_args ) );

    memset( &dparams, 0, sizeof( dparams ) );
    dparams.size = sizeof( dparams );
    memset( &status, 0, sizeof( status ) );
    status.size = sizeof( status );

    if( p_sys->in.numBufs == 0 || p_sys->out.numBufs == 0 )
    {
        if( InitXDMAudioDecoder( p_dec, &dparams ) == VLC_EGENERIC )
            goto error;
    }

    /* Setup input arguments */
    in_args.size = sizeof( in_args );
    in_args.numBytes = __MIN( p_sys->i_buffer + p_block->i_buffer,
                              (size_t)p_sys->in.bufSizes[0] );

    /* Setup input buffer */
    int i_used = in_args.numBytes - p_sys->i_buffer;
    assert( p_sys->i_buffer + p_block->i_buffer <= (size_t)p_sys->in.bufSizes[0] );

    if( i_used > 0 )
    {
        memcpy( p_sys->in.bufs[0] + p_sys->i_buffer, p_block->p_buffer, i_used );
        p_block->p_buffer += i_used;
        p_block->i_buffer -= i_used;
#ifdef TRACK_PTS
        p_sys->p_tracked_blocks[p_sys->i_tracked_blocks].i_pts = p_block->i_pts;
        p_sys->p_tracked_blocks[p_sys->i_tracked_blocks].i_left = i_used;
        p_sys->i_tracked_blocks++;
        assert( p_sys->i_tracked_blocks < 100 );
#endif
    }

    if( in_args.numBytes < p_sys->in.bufSizes[0] )
    {
        msg_Dbg( p_dec, "Waiting for more data" );
        block_Release( p_block );
        p_sys->i_buffer = in_args.numBytes;
        return NULL;
    }

    /* Setup output arguments */
    out_args.size = sizeof( out_args );

    /* Decode the audio */
    i = AUDDEC_process( p_sys->d, &p_sys->in, &p_sys->out, &in_args, &out_args );
    if( i != AUDDEC_EOK )
    {
        msg_Err( p_dec, "Audio decoding failed: %s",
                 davinci_GetExtendedError( out_args.extendedError ) );
        if( i == AUDDEC_ERUNTIME || out_args.bytesConsumed == 0 ||
            XDM_ISFATALERROR( out_args.extendedError ) )
        {
            goto error;
        }
        else
        {
            msg_Err( p_dec, "Rejecting frame" );
            b_reject = true;
        }
    }

    if( AUDDEC_control( p_sys->d, XDM_GETSTATUS, &dparams, &status ) != AUDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to get decoder status" );
        goto error;
    }

#ifdef DEBUG_DAVINCI
    msg_Dbg( p_dec, "Bytes consumed: %d", (int)out_args.bytesConsumed );
    msg_Dbg( p_dec, "Fast Fwd Len: %d", (int)status.fastFwdLen );
    msg_Dbg( p_dec, "autoPosition: %d", (int)status.autoPosition );
#endif

    p_sys->i_buffer = in_args.numBytes - out_args.bytesConsumed;
    memmove( p_sys->in.bufs[0], p_sys->in.bufs[0] + out_args.bytesConsumed,
             p_sys->i_buffer );
#ifdef TRACK_PTS
    mtime_t i_pts = p_sys->p_tracked_blocks[0].i_pts;

    i = 0;
    while( p_sys->p_tracked_blocks[i].i_left <= out_args.bytesConsumed )
    {
        out_args.bytesConsumed -= p_sys->p_tracked_blocks[i].i_left;
        i++;
    }
    p_sys->p_tracked_blocks[i].i_left -= out_args.bytesConsumed;
    if( i > 0 )
    {
        p_sys->i_tracked_blocks -= i;
        memmove( p_sys->p_tracked_blocks, p_sys->p_tracked_blocks+i, p_sys->i_tracked_blocks * sizeof( struct block_tracker_t ) );
    }
#else
    mtime_t i_pts = p_block->i_pts;
#endif

    if( !b_reject && p_dec->fmt_out.audio.i_rate != (unsigned int)status.sampleRate )
    {
        msg_Dbg( p_dec, "Setting audio output format info" );

        p_dec->fmt_out.audio.i_rate = status.sampleRate;
        msg_Dbg( p_dec, "Rate: %d", p_dec->fmt_out.audio.i_rate );
        aout_DateInit( &p_sys->date, p_dec->fmt_out.audio.i_rate );
        aout_DateSet( &p_sys->date, i_pts/*p_block->i_pts*/ );

        p_dec->fmt_out.audio.i_channels = status.numChannels;
        switch( status.numChannels )
        {
            case IAUDIO_MONO:
                i = AOUT_CHAN_CENTER;
                break;
            case IAUDIO_STEREO:
                i = AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT;
                break;
            case IAUDIO_THREE_ZERO:
                i = AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_LFE;
                break;
            case IAUDIO_FIVE_ZERO:
                i = AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT | AOUT_CHAN_CENTER ;
                break;
            case IAUDIO_FIVE_ONE:
                i = AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT | AOUT_CHAN_CENTER | AOUT_CHAN_LFE;
                break;
            case IAUDIO_SEVEN_ONE:
                i = AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_MIDDLELEFT | AOUT_CHAN_MIDDLERIGHT | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT | AOUT_CHAN_CENTER | AOUT_CHAN_LFE;
                break;
            default:
                msg_Warn( p_dec, "Unknown numChannels (%d)", (int)status.numChannels );
                i = status.numChannels;
                break;
        }
        p_dec->fmt_out.audio.i_physical_channels =
        p_dec->fmt_out.audio.i_original_channels = i;
        msg_Dbg( p_dec, "Channels: %d", p_dec->fmt_out.audio.i_original_channels );

        p_dec->fmt_out.audio.i_bitspersample = status.outputBitsPerSample;
        msg_Dbg( p_dec, "Bits per sample: %d", p_dec->fmt_out.audio.i_bitspersample );

        assert( p_dec->fmt_out.audio.i_bitspersample == 16 );
    }

    if( i_pts != 0 && i_pts /*p_block->i_pts != 0 && p_block->i_pts*/ != aout_DateGet( &p_sys->date ) )
    {
        aout_DateSet( &p_sys->date, i_pts/* p_block->i_pts */);
    }
    else if( !aout_DateGet( &p_sys->date ) )
    {
        /* We're still waiting for a pts */
        goto error;
    }

    p_out = decoder_NewAudioBuffer( p_dec, b_reject ? 0 : status.frameLen /
                                              p_dec->fmt_out.audio.i_channels );
    if( !p_out )
    {
        block_Release( p_block );
        p_sys->i_buffer = 0;
        return NULL;
    }

    p_out->start_date = aout_DateGet( &p_sys->date );
    p_out->end_date = aout_DateIncrement( &p_sys->date, b_reject ?
                        0 : status.frameLen / p_dec->fmt_out.audio.i_channels );
    memcpy( p_out->p_buffer, p_sys->out.bufs[0], p_out->i_nb_bytes );

    return p_out;

error:
    if( p_out )
        decoder_DeleteAudioBuffer( p_dec, p_out );
    block_Release( p_block );
    p_sys->i_buffer = 0;
    return NULL;
}
