/*****************************************************************************
 * encoder.c: video encoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2010 M2X BV
 * $Id: d821f87ae0f108474af388325d2c49029d4cdc07 $
 *
 * Authors: Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_vout.h>
#include <vlc_aout.h>
#include <vlc_sout.h>
#include <vlc_codec.h>

#include "davinci.h"

/* XDM 1.0 */
#include <ti/sdo/ce/audio1/audenc1.h>

/* AAC LC Encoder */
#include <ti/sdo/ce/audio1/iaacenc.h>

#include <assert.h>

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/

/* XDM 1.0 */
static block_t *EncodeAudio1( encoder_t *, aout_buffer_t * );

static void davinci_get_audio_channels( int i_physical_channels, int i_original_channels,
        IAUDIO_ChannelMode *channel_mode, IAUDIO_DualMonoMode *dual_mono, XDAS_Int32 *lfe );

/*****************************************************************************
 * Encoder
 *****************************************************************************/

struct xdm_sys_t
{
    /* XDM 1.0 */
    AUDENC1_Handle handle;

    AUDENC1_Params params;
    AUDENC1_DynamicParams dparams;

    AUDENC1_InArgs in_args;
    XDM1_BufDesc in;

    AUDENC1_OutArgs out_args;
    XDM1_BufDesc out;

    /* misc */
    uint32_t i_channels;
    uint32_t i_frames; /* number of sample frames needed */
    uint32_t i_bytespersample;

    /* remainder buffer management */
    uint8_t  *p_tmp;
    size_t   i_tmp;

    size_t   i_nb_bytes;
    uint32_t i_nb_samples;
    mtime_t  start_date;
    mtime_t  end_date;
};

/*****************************************************************************
 *
 *****************************************************************************/
int OpenEncoderAudio1( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm;

    p_sys->xdm = xdm = (xdm_sys_t *) calloc( 1, sizeof( xdm_sys_t ) );
    if( !p_sys->xdm )
         return VLC_ENOMEM;

    /* */
    xdm->p_tmp = NULL;
    xdm->i_channels = aout_FormatNbChannels(&p_enc->fmt_out.audio) ?
                              aout_FormatNbChannels(&p_enc->fmt_out.audio) :
                              p_enc->fmt_out.audio.i_channels;

    xdm->i_bytespersample = p_enc->fmt_out.audio.i_bitspersample >> 3;

    /* */
    xdm->params.size = sizeof( xdm->params );
    xdm->dparams.size = sizeof( xdm->dparams );

    memset( &xdm->in_args, 0, sizeof( xdm->in_args ) );
    xdm->in_args.size = sizeof( xdm->in_args );
    xdm->in_args.ancData.bufSize = 0L;
    xdm->in_args.ancData.buf     = (XDAS_Int8 *)NULL;

    memset( &xdm->out_args, 0, sizeof( xdm->out_args ) );
    xdm->out_args.size = sizeof( xdm->out_args );

    /* */
    p_enc->i_tolerance = var_CreateGetInteger( p_enc, ENC_CFG_PREFIX "tolerance" );
    p_enc->i_tolerance *= 1024; /* bits per second */

    xdm->params.encMode = var_CreateGetBool( p_enc, ENC_CFG_PREFIX "cbr" ) ?
                                             IAUDIO_CBR : IAUDIO_VBR;
    xdm->params.bitRate = p_enc->fmt_out.i_bitrate;
    if( xdm->params.encMode == IAUDIO_VBR )
        xdm->params.maxBitRate = p_enc->fmt_out.i_bitrate + p_enc->i_tolerance;
    else
        xdm->params.maxBitRate = p_enc->fmt_out.i_bitrate;
    xdm->params.sampleRate = p_enc->fmt_out.audio.i_rate;
    davinci_get_audio_channels( p_enc->fmt_out.audio.i_physical_channels,
                                p_enc->fmt_out.audio.i_original_channels,
                                (IAUDIO_ChannelMode*)&xdm->params.channelMode,
                                (IAUDIO_DualMonoMode*)&xdm->params.dualMonoMode,
                                &xdm->params.lfeFlag );
    xdm->params.dataEndianness = XDM_LE_16;
    xdm->params.inputFormat = IAUDIO_INTERLEAVED; /* IAUDIO_BLOCK; */
    xdm->params.inputBitsPerSample = p_enc->fmt_out.audio.i_bitspersample;
    xdm->params.crcFlag = XDAS_FALSE;
    xdm->params.ancFlag = XDAS_FALSE;

    msg_Info( p_enc, "DSP algorithm '%s', audio %d channels, samplerate %d Hz (max bitrate %d kbps)\n",
        p_sys->psz_codec, xdm->i_channels,
        (int)xdm->params.sampleRate,
        ((int)xdm->params.maxBitRate/1000) );

    /* Set codec specific parameters */
    if( VLC_FOURCC('m','p','4','a') == p_enc->fmt_out.i_codec ||
        VLC_FOURCC('a','a','c',' ') == p_enc->fmt_out.i_codec )
    {
        IAACENC_Params aacParams;
        char *psz_aac_profile = var_CreateGetString( p_enc, ENC_CFG_PREFIX "aac-profile" );

        xdm->params.size = sizeof( aacParams );
        memcpy( &aacParams.audenc_params, &(xdm->params), sizeof( xdm->params ) );

        xdm->i_frames = 2048;                                 /* HE AAC frame size */
        if( !strncmp(psz_aac_profile, "sbr", 3) )
            aacParams.outObjectType = AACENC_OBJ_TYP_HEAAC;     /* HE-AAC with SBR */
        else if( !strncmp(psz_aac_profile, "ps", 2) )
            aacParams.outObjectType = AACENC_OBJ_TYP_PS; /* HE-AAC with SBR and PS */
        else if( !strncmp(psz_aac_profile, "low", 3) )
        {
            aacParams.outObjectType = AACENC_OBJ_TYP_LC;     /* Low Complexity AAC */
            xdm->i_frames = 1024;                             /* LC AAC frame size */
        }
        else
        {
            msg_Err( p_enc, "AAC encoder does not support this profile: '%s'",
                     psz_aac_profile );
            free( psz_aac_profile );
            goto error;
        }
        free( psz_aac_profile );

        msg_Dbg( p_enc, "using aac-profile '%s'",
                 aacParams.outObjectType == AACENC_OBJ_TYP_HEAAC ? "sbr" :
                     aacParams.outObjectType == AACENC_OBJ_TYP_PS ? "ps" : "low" );

        /* boundary checks */
        switch( aacParams.outObjectType )
        {
            default:
            case AACENC_OBJ_TYP_LC:
                if( xdm->params.maxBitRate > (192*1024) ||
                    xdm->params.sampleRate > 96000 ||
                    xdm->params.sampleRate <  8000 )
                {
                    msg_Err( p_enc, "samplerate between range 8000 and 96000 Hz and "
                                    "bitrate smaller then 192 kbit/s expected" );
                    goto error;
                }
                if( xdm->params.channelMode != IAUDIO_2_0 )
                {
                    msg_Err( p_enc,
                             "AAC LC encoder only supports 2 channels %d requested",
                             xdm->i_channels );
                    goto error;
                }
                break;
            case AACENC_OBJ_TYP_HEAAC:
            case AACENC_OBJ_TYP_PS:
                if( xdm->params.maxBitRate > (64*1024) ||
                    xdm->params.sampleRate > 48000 ||
                    xdm->params.sampleRate < 16000 )
                {
                    msg_Err( p_enc, "samplerate between range 16000 and 48000 Hz and "
                                    "bitrate smaller then 64 kbit/s expected" );
                    goto error;
                }
                break;
        }

        aacParams.outFileFormat = AACENC_TT_ADTS;    /* ADTS formatted */
        aacParams.useTns = AACENC_TRUE;
        aacParams.usePns = AACENC_TRUE;
        aacParams.downMixFlag = AACENC_FALSE;
        aacParams.bitRateMode = AACENC_BR_MODE_VBR_5;
        aacParams.ancRate = 0; /* default -1 */

        xdm->handle = AUDENC1_create( p_sys->engine, (String)p_sys->psz_codec,
                                      (AUDENC1_Params *)&aacParams );
    }
    else
    {
        xdm->handle = AUDENC1_create( p_sys->engine, (String)p_sys->psz_codec,
                                      &xdm->params );
    }

    if( !xdm->handle )
    {
        goto error;
    }

    p_sys->pf_encode_audio = EncodeAudio1;
    return VLC_SUCCESS;

error:
    free( p_sys->xdm );
    p_sys->xdm = NULL;
    return VLC_EGENERIC;
}

void CloseEncoderAudio1( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm = p_sys->xdm;

    if( !xdm ) return;

    /* Delete audio encoder */
    if( p_enc->fmt_out.i_cat == AUDIO_ES )
        AUDENC1_delete( xdm->handle );

    /* Free 'DaVinci compliant' buffers */
    davinci_FreeBuffer1( &xdm->in );
    davinci_FreeBuffer1( &xdm->out );

    free( xdm->p_tmp );
    free( xdm );
    xdm = NULL;
}

/*****************************************************************************
 * Audio Encoder: XDM 1.0 API
 *****************************************************************************/
static void davinci_get_audio_channels( int i_physical_channels, int i_original_channels,
                    IAUDIO_ChannelMode *channel_mode, IAUDIO_DualMonoMode *dual_mono, XDAS_Int32 *lfe )
{
    /* default setting */
    *channel_mode = IAUDIO_2_0;
    *dual_mono = IAUDIO_DUALMONO_LR;
    *lfe = XDAS_FALSE;

    switch ( i_physical_channels & AOUT_CHAN_PHYSMASK )
    {
    case AOUT_CHAN_LEFT:
    case AOUT_CHAN_RIGHT:
    case AOUT_CHAN_CENTER:
        if ( (i_original_channels & AOUT_CHAN_CENTER)
              || (i_original_channels
                   & (AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT)) )
            *channel_mode = IAUDIO_1_0;
        else if ( i_original_channels & AOUT_CHAN_LEFT )
            *channel_mode = IAUDIO_1_0;
        else
            *channel_mode = IAUDIO_1_0;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT:
        if ( i_original_channels & AOUT_CHAN_REVERSESTEREO )
        {
            if ( i_original_channels & AOUT_CHAN_DOLBYSTEREO )
                *channel_mode = IAUDIO_2_0;
            else
                *channel_mode = IAUDIO_2_0;
        }
        else
        {
            if ( i_original_channels & AOUT_CHAN_DUALMONO )
            {
                *channel_mode = IAUDIO_11_0;
                *dual_mono = IAUDIO_DUALMONO_LR;
            }
            else if ( i_original_channels == AOUT_CHAN_CENTER )
                *channel_mode = IAUDIO_2_0;
            else if ( !(i_original_channels & AOUT_CHAN_RIGHT) )
                *channel_mode = IAUDIO_2_0;
            else if ( !(i_original_channels & AOUT_CHAN_LEFT) )
                *channel_mode = IAUDIO_2_0;
            else
                *channel_mode = IAUDIO_2_0;
        }
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER:
        *channel_mode = IAUDIO_3_0;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_REARCENTER:
        *channel_mode = IAUDIO_2_1;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER
          | AOUT_CHAN_REARCENTER:
        *channel_mode = IAUDIO_3_1;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT
          | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT:
        *channel_mode = IAUDIO_2_2;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT
          | AOUT_CHAN_MIDDLELEFT | AOUT_CHAN_MIDDLERIGHT:
        *channel_mode = IAUDIO_2_2;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER
          | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT:
        *channel_mode = IAUDIO_3_2;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER
          | AOUT_CHAN_MIDDLELEFT | AOUT_CHAN_MIDDLERIGHT:
        *channel_mode = IAUDIO_3_3;
        break;
    /* LFE */
    case AOUT_CHAN_CENTER | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        if ( (i_original_channels & AOUT_CHAN_CENTER)
              || (i_original_channels
                   & (AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT)) )
            *channel_mode = IAUDIO_1_0;
        else if ( i_original_channels & AOUT_CHAN_LEFT )
            *channel_mode = IAUDIO_1_0;
        else
            *channel_mode = IAUDIO_1_0;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        if ( i_original_channels & AOUT_CHAN_DOLBYSTEREO )
            *channel_mode = IAUDIO_2_0;
        else if ( i_original_channels & AOUT_CHAN_DUALMONO )
        {
            *channel_mode = IAUDIO_11_0;
            *dual_mono = IAUDIO_DUALMONO_LR;
        }
        else if ( i_original_channels == AOUT_CHAN_CENTER )
            *channel_mode = IAUDIO_1_0;
        else if ( !(i_original_channels & AOUT_CHAN_RIGHT) )
            *channel_mode = IAUDIO_2_0;
        else if ( !(i_original_channels & AOUT_CHAN_LEFT) )
            *channel_mode = IAUDIO_2_0;
        else
            *channel_mode = IAUDIO_2_0;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        *channel_mode = IAUDIO_3_0;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_REARCENTER
          | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        *channel_mode = IAUDIO_2_1;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER
          | AOUT_CHAN_REARCENTER | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        *channel_mode = IAUDIO_3_1;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT
          | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        *channel_mode = IAUDIO_2_2;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT
          | AOUT_CHAN_MIDDLELEFT | AOUT_CHAN_MIDDLERIGHT | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        *channel_mode = IAUDIO_2_3;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER
          | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        *channel_mode = IAUDIO_3_2;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER
          | AOUT_CHAN_MIDDLELEFT | AOUT_CHAN_MIDDLERIGHT | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        *channel_mode = IAUDIO_3_3;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER
          | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT | AOUT_CHAN_MIDDLELEFT
          | AOUT_CHAN_MIDDLERIGHT:
        *channel_mode = IAUDIO_3_4;
        break;
    case AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT | AOUT_CHAN_CENTER
          | AOUT_CHAN_REARLEFT | AOUT_CHAN_REARRIGHT | AOUT_CHAN_MIDDLELEFT
          | AOUT_CHAN_MIDDLERIGHT | AOUT_CHAN_LFE:
        *lfe = XDAS_TRUE;
        *channel_mode = IAUDIO_3_4;
        break;
    }
}

/* NOTE: The DSP TI audio codecs generate a output packet with header and
 * 0 as data on a parameters change. Thus only change parameters when it is
 * needed with davinci_SetDynamicParams.
 */
static inline int davinci_SetDynamicParams( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm = p_sys->xdm;
    AUDENC1_Status status;

    memset( &status, 0, sizeof( status ) );
    status.size = sizeof( status );

    xdm->dparams.bitRate = p_enc->fmt_out.i_bitrate;
    xdm->dparams.sampleRate = p_enc->fmt_out.audio.i_rate;
    davinci_get_audio_channels( p_enc->fmt_out.audio.i_physical_channels,
                                p_enc->fmt_out.audio.i_original_channels,
                                (IAUDIO_ChannelMode*)&xdm->dparams.channelMode,
                                (IAUDIO_DualMonoMode*)&xdm->dparams.dualMonoMode,
                                &xdm->dparams.lfeFlag );
    assert( xdm->dparams.channelMode == xdm->params.channelMode );

    xdm->dparams.inputBitsPerSample = p_enc->fmt_out.audio.i_bitspersample;

    if( AUDENC1_control( xdm->handle, XDM_SETPARAMS, &xdm->dparams, &status )
        != AUDENC1_EOK )
    {
        msg_Err( p_enc, "Failed to set encoder parameters" );
        return VLC_EGENERIC;
    }

    return VLC_SUCCESS;
}

static inline int davinci_InitAudio1Buffers( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm = p_sys->xdm;
    int i_ret = VLC_SUCCESS;

    AUDENC1_Status status;

    memset( &status, 0, sizeof( status ) );
    status.size = sizeof( status );

    /* Configure buffers */
    if( AUDENC1_control( xdm->handle, XDM_GETBUFINFO, &xdm->dparams, &status )
        != AUDENC1_EOK )
    {
        msg_Err( p_enc, "Failed to get buffer info" );
        return VLC_EGENERIC;
    }

    /* Allocate input buffer(s) */
    if( (i_ret = davinci_AllocateBuffer1( status.bufInfo.minNumInBufs,
                         status.bufInfo.minInBufSize, &xdm->in ))
        != VLC_SUCCESS )
    {
        msg_Err( p_enc, "Failed to allocate input buffers" );
        return i_ret;
    }

    /* Allocate output buffer */
    if( (i_ret = davinci_AllocateBuffer1( status.bufInfo.minNumOutBufs,
                                  status.bufInfo.minOutBufSize, &xdm->out ))
        != VLC_SUCCESS )
    {
        msg_Err( p_enc, "Failed to allocate output buffers" );
        return i_ret;
    }

    return i_ret;
}
#if 0
static void DumpBuffers( const uint8_t *src, const uint32_t size )
{
    int n = 0;
    printf("\n");
    for( uint32_t i = 0; i < size; i++, n++ )
    {
        if ((n % 20) == 0)
           printf( "\n %.4d: ", n );
        printf( "0x%.2x ", src[i] );
    }
    printf("\n");
}
#endif
static block_t *EncodeAudio1( encoder_t *p_enc, aout_buffer_t *p_buffer )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm = p_sys->xdm;
    block_t *p_block;
    int i_ret;

    uint32_t i_nb_samples = 0;
    mtime_t i_pts = (mtime_t)0;
    mtime_t i_pts_end = (mtime_t)0;

    if( xdm->in.numBufs == 0 || xdm->out.numBufs == 0 )
    {
        if( davinci_InitAudio1Buffers( p_enc ) != VLC_SUCCESS )
            return NULL;

        if( xdm->p_tmp == NULL )
        {
            xdm->p_tmp = (int8_t *) malloc( (uint32_t)xdm->in.descs[0].bufSize * sizeof(int8_t) );
            if( xdm->p_tmp == NULL )
                return NULL;
        }

        if( davinci_SetDynamicParams( p_enc ) != VLC_SUCCESS )
        {
            msg_Warn( p_enc, "Failed setting dynamic params" );
        }

        msg_Info( p_enc, "using %d channels at %d Hz samplerate (bitrate %d kbp/s, "
                         "%d bits per sample)\n",
                         xdm->i_channels,
                         (int)xdm->dparams.sampleRate,
                         (int)xdm->dparams.bitRate/1000,
                         (int)xdm->dparams.inputBitsPerSample );

    }

    /* buffer management */
    if( (p_buffer->i_nb_bytes == (size_t)xdm->in.descs[0].bufSize) &&
        (xdm->i_nb_bytes == 0) )
    {
        /* buffer fits in encoder buffer entirely */
        memcpy( (uint8_t*)xdm->in.descs[0].buf, p_buffer->p_buffer, p_buffer->i_nb_bytes );

        /* samples in DSP buffer */
        i_nb_samples = p_buffer->i_nb_samples;

        i_pts = p_buffer->start_date;
        i_pts_end = p_buffer->end_date;

        xdm->start_date = 0;
        xdm->end_date = 0;
    }
    else
    {
        uint32_t i_nb_delayed_samples = 0;
        size_t i_size = 0;

        /* delayed pts */
        if( xdm->start_date == 0 )
            xdm->start_date = p_buffer->start_date;

        if( xdm->i_tmp > 0 )
        {
            memcpy( (uint8_t*)xdm->in.descs[0].buf + xdm->i_nb_bytes, xdm->p_tmp, xdm->i_tmp );
            xdm->i_nb_bytes += xdm->i_tmp;
            xdm->i_tmp = 0;
        }

        if( xdm->i_nb_bytes + p_buffer->i_nb_bytes <= (size_t)xdm->in.descs[0].bufSize )
        {
            memcpy( (uint8_t*)xdm->in.descs[0].buf + xdm->i_nb_bytes, p_buffer->p_buffer, p_buffer->i_nb_bytes );

            xdm->i_nb_bytes += p_buffer->i_nb_bytes;
            xdm->i_nb_samples += p_buffer->i_nb_samples;

            xdm->end_date = p_buffer->end_date;

            if( xdm->i_nb_bytes < (size_t)xdm->in.descs[0].bufSize )
            {
                /* wait for more samples to arrive */
                return NULL;
            }
        }
        else // if( xdm->i_nb_bytes + p_buffer->i_nb_bytes > (size_t)xdm->in.descs[0].bufSize )
        {
            i_size = xdm->in.descs[0].bufSize - xdm->i_nb_bytes;
            i_size = __MIN(i_size, p_buffer->i_nb_bytes);
            memcpy( (uint8_t*)xdm->in.descs[0].buf + xdm->i_nb_bytes, p_buffer->p_buffer, i_size );

            i_nb_delayed_samples = p_buffer->i_nb_samples - (xdm->i_frames - xdm->i_nb_samples);
            xdm->i_nb_samples += (p_buffer->i_nb_samples - i_nb_delayed_samples);
            xdm->i_nb_bytes += i_size;

            /* compensate for incomplete buffer copy */
            xdm->end_date = p_buffer->start_date +
                        ((p_buffer->end_date - p_buffer->start_date) /
                          p_buffer->i_nb_samples) *
                          (p_buffer->i_nb_samples - i_nb_delayed_samples);
        }

        /* always fill entire encoder bufer */
        assert( (uint32_t)xdm->in.descs[0].bufSize == xdm->i_nb_bytes );

        /* samples in DSP buffer */
        i_nb_samples = xdm->i_nb_samples;

        i_pts = xdm->start_date;
        i_pts_end = xdm->end_date;

        /* keep remainder */
        if( i_nb_delayed_samples > 0 )
        {
            memset( xdm->p_tmp, 0, (uint32_t)xdm->in.descs[0].bufSize );
            memcpy( xdm->p_tmp, p_buffer->p_buffer + i_size, p_buffer->i_nb_bytes - i_size );
            xdm->i_tmp = p_buffer->i_nb_bytes - i_size;

            xdm->i_nb_bytes = 0;
            xdm->i_nb_samples = i_nb_delayed_samples;

            xdm->start_date = i_pts_end;
            xdm->end_date = p_buffer->end_date;
        }
        else
        {
            xdm->start_date = xdm->end_date = (mtime_t) 0;
            xdm->i_nb_bytes = xdm->i_nb_samples = 0;
        }
    }

    /* test: for max samples */
    assert( i_nb_samples == xdm->i_frames );

    /* nr. of input samples */
    xdm->in_args.numInSamples = i_nb_samples;

    /* Encode the audio */
    i_ret = AUDENC1_process( xdm->handle, (XDM1_BufDesc *)&xdm->in,
                             (XDM1_BufDesc *)&xdm->out,
                             &xdm->in_args, &xdm->out_args );
    if( i_ret != AUDENC1_EOK )
    {
        msg_Err( p_enc, "Audio encoding failed (%d): %s", i_ret,
                 davinci_GetExtendedError( xdm->out_args.extendedError ) );
        return NULL;
    }

    assert( xdm->out_args.extendedError == 0 ); // assumption - test code

#if 0
    /* Print some info */
    msg_Dbg( p_enc, "Bytes generated: %d, buffer: %d/%d, samples: %d/%d, zeroed: %d",
              (int)xdm->out_args.bytesGenerated,
              (int)xdm->in.descs[0].bufSize, (int)xdm->out.descs[0].bufSize,
              (int)xdm->in_args.numInSamples, (int)xdm->out_args.numInSamples,
              (int)xdm->out_args.numZeroesPadded );
#endif

    /* Put everything in the block */
    if( xdm->out_args.bytesGenerated <= 0 )
        return NULL;

    /* process samples */
    p_block = block_New( p_enc, xdm->out_args.bytesGenerated );
    if( !p_block )
        return NULL;

    memcpy( p_block->p_buffer, (uint8_t *)(xdm->out.descs[0].buf), p_block->i_buffer );

#if 0 /* output buffer */
    DumpBuffers( p_block->p_buffer, (uint32_t) xdm->out_args.bytesGenerated );
#endif

    p_block->i_length = i_pts_end - i_pts;
    p_block->i_dts = p_block->i_pts = i_pts;

    return p_block;
}
