/*****************************************************************************
 * encoder.c: video encoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008-2010 M2X BV
 * $Id: 03a0238ae4521d344749eca7d32aaf68bba8b98c $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *          Rafaël Carré <rcarre@m2x.nl>
 *          Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_vout.h>
#include <vlc_aout.h>
#include <vlc_sout.h>
#include <vlc_codec.h>

#include "davinci.h"

/* XDM 0.9 */
#include <ti/sdo/ce/audio/audenc.h>

#include <assert.h>

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
/* XDM 0.9 */
static block_t *EncodeAudio( encoder_t *, aout_buffer_t * );

/*****************************************************************************
 * Encoder
 *****************************************************************************/

struct xdm_sys_t
{
    /* XDM 0.9 */
    AUDENC_Handle handle;

    AUDENC_Params params;
    AUDENC_DynamicParams dparams;

    AUDENC_InArgs in_args;
    XDM_BufDesc in;

    AUDENC_OutArgs out_args;
    XDM_BufDesc out;

    /* misc */
    uint32_t i_channels;
    uint32_t i_frames; /* number of sample frames needed */
    uint32_t i_bytespersample;

    /* remainder buffer management */
    uint8_t  *p_tmp;
    size_t   i_tmp;

    size_t   i_nb_bytes;
    uint32_t i_nb_samples;
    mtime_t  start_date;
    mtime_t  end_date;

};

/*****************************************************************************
 *
 *****************************************************************************/

int OpenEncoderAudio( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm;

    xdm = (xdm_sys_t *) calloc( 1, sizeof( xdm_sys_t ) );
    if( !xdm )
         return VLC_ENOMEM;
    xdm = p_sys->xdm;

    /* */
    xdm->p_tmp = NULL;
    xdm->i_channels = aout_FormatNbChannels(&p_enc->fmt_out.audio) ?
                              aout_FormatNbChannels(&p_enc->fmt_out.audio) :
                              p_enc->fmt_out.audio.i_channels;

    xdm->i_bytespersample = p_enc->fmt_out.audio.i_bitspersample >> 3;

    /* */
    xdm->params.size = sizeof( xdm->params );
    xdm->dparams.size = sizeof( xdm->dparams );

    memset( &xdm->in_args, 0, sizeof( xdm->in_args ) );
    xdm->in_args.size = sizeof( xdm->in_args );

    memset( &xdm->out_args, 0, sizeof( xdm->out_args ) );
    xdm->out_args.size = sizeof( xdm->out_args );

    switch( var_CreateGetInteger( p_enc, ENC_CFG_PREFIX "quality" ) )
    {
        case 1: xdm->params.encodingPreset = XDM_HIGH_QUALITY; break;
        case 2: xdm->params.encodingPreset = XDM_HIGH_SPEED; break;
        case 3: xdm->params.encodingPreset = XDM_USER_DEFINED; break;
        default:
            xdm->params.encodingPreset = XDM_DEFAULT;
            break;
    }
    xdm->params.maxBitrate = p_enc->fmt_out.i_bitrate;
    xdm->params.maxSampleRate = p_enc->fmt_out.audio.i_rate;
    xdm->params.maxNoOfCh = p_enc->fmt_out.audio.i_channels;
    xdm->params.dataEndianness = XDM_BYTE;

    msg_Info( p_enc, "DSP algorithm '%s', audio %d channels, %d Hz (max bitrate %d kb/s)\n",
        p_sys->psz_codec,
        (int)xdm->params.maxNoOfCh,
        (int)xdm->params.maxSampleRate,
        ((int)xdm->params.maxBitrate)/1000 );

    /* Create encoder handle */
    xdm->handle = AUDENC_create( p_sys->engine, (String)p_sys->psz_codec, &xdm->params );
    if( !xdm->handle )
        goto error;

    p_sys->pf_encode_audio = EncodeAudio;
    return VLC_SUCCESS;

error:
    free( p_sys->xdm );
    p_sys->xdm = NULL;
    return VLC_EGENERIC;
}

void CloseEncoderAudio( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm = p_sys->xdm;

    if( !xdm ) return;

    /* Delete audio encoder */
    if( p_enc->fmt_out.i_cat == AUDIO_ES )
        AUDENC_delete( xdm->handle );

    /* Free 'DaVinci compliant' buffers */
    davinci_FreeBuffer( &xdm->in );
    davinci_FreeBuffer( &xdm->out );

    free( xdm->p_tmp );
    free( xdm );
    xdm = NULL;
}

/*****************************************************************************
 * Audio Encoder: XDM 0.9
 *****************************************************************************/

/* NOTE: The DSP TI audio codecs generate a output packet with header and
 * 0 as data on a parameters change. Thus only change parameters when it is
 * needed with davinci_SetDynamicParams.
 */
static inline int davinci_SetDynamicParams( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm = p_sys->xdm;
    AUDENC_Status status;

    memset( &status, 0, sizeof( status ) );
    status.size = sizeof( status );

    xdm->dparams.inputFormat = IAUDIO_INTERLEAVED; // IAUDIO_BLOCK;
    xdm->dparams.bitRate = p_enc->fmt_out.i_bitrate;
    xdm->dparams.sampleRate = p_enc->fmt_out.audio.i_rate;
    switch( p_enc->fmt_out.audio.i_channels )
    {
        case 1: xdm->dparams.numChannels = IAUDIO_MONO; /* Single channel. */
            break;
        case 2: xdm->dparams.numChannels = IAUDIO_STEREO; /* Two channel. */
            break;
        case 3: xdm->dparams.numChannels = IAUDIO_THREE_ZERO; /* Three channel. */
            break;
        case 5: xdm->dparams.numChannels = IAUDIO_FIVE_ZERO; /* Five channel. */
            break;
        case 6: xdm->dparams.numChannels = IAUDIO_FIVE_ONE; /* 5.1 channel. */
            break;
        case 7: xdm->dparams.numChannels = IAUDIO_SEVEN_ONE; /* 7.1 channel. */
            break;
        default:
            msg_Warn( p_enc, "assuming stereo audio channels" );
            xdm->dparams.numChannels = IAUDIO_STEREO; /**< Two channel. */
            break;
    }
    xdm->dparams.numLFEChannels = 0;  /* FIXME: What to fill in here?
         * Number of LFE (Low Frequency Effects) channels in the stream. */
    xdm->dparams.inputBitsPerSample = p_enc->fmt_out.audio.i_bitspersample;

    if( AUDENC_control( xdm->handle, XDM_SETPARAMS, &xdm->dparams, &status )
        != AUDENC_EOK )
    {
        msg_Err( p_enc, "Failed to set encoder parameters." );
        return VLC_EGENERIC;
    }

    return VLC_SUCCESS;
}

static inline int davinci_InitAudioBuffers( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm = p_sys->xdm;
    int i_ret = VLC_SUCCESS;

    AUDENC_Status status;

    memset( &status, 0, sizeof( status ) );
    status.size = sizeof( status );

    /* Configure buffers */
    if( AUDENC_control( xdm->handle, XDM_GETBUFINFO, &xdm->dparams, &status )
        != AUDENC_EOK )
    {
        msg_Err( p_enc, "Failed to get buffer info" );
        return VLC_EGENERIC;
    }

    xdm->i_frames = status.frameLen;

    /* Allocate input buffer(s) */
    if( (i_ret = davinci_AllocateBuffer( status.bufInfo.minNumInBufs,
                                  status.bufInfo.minInBufSize, &xdm->in ))
        != VLC_SUCCESS )
    {
        msg_Err( p_enc, "Failed to allocate input buffers" );
        return i_ret;
    }

    /* Allocate output buffer */
    if( (i_ret = davinci_AllocateBuffer( status.bufInfo.minNumOutBufs,
                                  status.bufInfo.minOutBufSize, &xdm->out ))
        != VLC_SUCCESS )
    {
        msg_Err( p_enc, "Failed to allocate output buffers" );
        return i_ret;
    }

    return i_ret;
}

static block_t *EncodeAudio( encoder_t *p_enc, aout_buffer_t *p_buffer )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    xdm_sys_t *xdm = p_sys->xdm;
    block_t *p_block;
    int i_ret;

    uint32_t i_nb_samples = 0;
    mtime_t i_pts = (mtime_t)0;
    mtime_t i_pts_end = (mtime_t)0;

    if( xdm->in.numBufs == 0 || xdm->out.numBufs == 0 )
    {
        if( davinci_InitAudioBuffers( p_enc ) != VLC_SUCCESS )
            return NULL;

        if( xdm->p_tmp == NULL )
        {
            xdm->p_tmp = (int8_t *) malloc( (uint32_t)xdm->in.bufSizes[0] * sizeof(int8_t) );
            if( xdm->p_tmp == NULL )
                return NULL;
        }

        if( davinci_SetDynamicParams( p_enc ) != VLC_SUCCESS )
        {
            msg_Warn( p_enc, "Failed setting dynamic params." );
        }

         msg_Info( p_enc, "using %d channels at %d Hz samplerate (bitrate %d kBps, "
                     "%d bits per sample)\n",
                     p_enc->fmt_out.audio.i_channels,
                     (int)xdm->dparams.sampleRate,
                     ((int)xdm->dparams.bitRate) >> 13 /* / (8*1024)*/,
                     (int)xdm->dparams.inputBitsPerSample );
    }

    /* buffer management */
    if( (p_buffer->i_nb_bytes == (size_t)xdm->in.bufSizes[0]) &&
        (xdm->i_nb_bytes == 0) )
    {
        /* buffer fits in encoder buffer entirely */
        memcpy( (uint8_t*)xdm->in.bufs[0], p_buffer->p_buffer, p_buffer->i_nb_bytes );

        /* samples in DSP buffer */
        i_nb_samples = p_buffer->i_nb_samples;

        i_pts = p_buffer->start_date;
        i_pts_end = p_buffer->end_date;

        xdm->start_date = 0;
        xdm->end_date = 0;
    }
    else
    {
        uint32_t i_nb_delayed_samples = 0;
        size_t i_size = 0;

        /* delayed pts */
        if( xdm->start_date == 0 )
            xdm->start_date = p_buffer->start_date;

        if( xdm->i_tmp > 0 )
        {
            memcpy( (uint8_t*)xdm->in.bufs[0] + xdm->i_nb_bytes, xdm->p_tmp, xdm->i_tmp );
            xdm->i_nb_bytes += xdm->i_tmp;
            xdm->i_tmp = 0;
        }

        if( xdm->i_nb_bytes + p_buffer->i_nb_bytes <= (size_t)xdm->in.bufSizes[0] )
        {
            memcpy( (uint8_t*)xdm->in.bufs[0] + xdm->i_nb_bytes, p_buffer->p_buffer, p_buffer->i_nb_bytes );

            xdm->i_nb_bytes += p_buffer->i_nb_bytes;
            xdm->i_nb_samples += p_buffer->i_nb_samples;

            xdm->end_date = p_buffer->end_date;

            if( xdm->i_nb_bytes < (size_t)xdm->in.bufSizes[0] )
            {
                /* wait for more samples to arrive */
                return NULL;
            }
        }
        else // if( xdm->i_nb_bytes + p_buffer->i_nb_bytes > (size_t)xdm->in.bufSizes[0] )
        {
            i_size = xdm->in.bufSizes[0] - xdm->i_nb_bytes;
            i_size = __MIN(i_size, p_buffer->i_nb_bytes);
            memcpy( (uint8_t*)xdm->in.bufs[0] + xdm->i_nb_bytes, p_buffer->p_buffer, i_size );

            i_nb_delayed_samples = p_buffer->i_nb_samples - (xdm->i_frames - xdm->i_nb_samples);
            xdm->i_nb_samples += (p_buffer->i_nb_samples - i_nb_delayed_samples);
            xdm->i_nb_bytes += i_size;

            /* compensate for incomplete buffer copy */
            xdm->end_date = p_buffer->start_date +
                        ((p_buffer->end_date - p_buffer->start_date) /
                          p_buffer->i_nb_samples) *
                          (p_buffer->i_nb_samples - i_nb_delayed_samples);
        }

        /* always fill entire encoder bufer */
        assert( (uint32_t)xdm->in.bufSizes[0] == xdm->i_nb_bytes );

        /* samples in DSP buffer */
        i_nb_samples = xdm->i_nb_samples;

        i_pts = xdm->start_date;
        i_pts_end = xdm->end_date;

        /* keep remainder */
        if( i_nb_delayed_samples > 0 )
        {
            memset( xdm->p_tmp, 0, (uint32_t)xdm->in.bufSizes[0] );
            memcpy( xdm->p_tmp, p_buffer->p_buffer + i_size, p_buffer->i_nb_bytes - i_size );
            xdm->i_tmp = p_buffer->i_nb_bytes - i_size;

            xdm->i_nb_bytes = 0;
            xdm->i_nb_samples = i_nb_delayed_samples;

            xdm->start_date = i_pts_end;
            xdm->end_date = p_buffer->end_date;
        }
        else
        {
            xdm->start_date = xdm->end_date = (mtime_t) 0;
            xdm->i_nb_bytes = xdm->i_nb_samples = 0;
        }
    }

    /* test: for max samples */
    assert( i_nb_samples == xdm->i_frames );

    /* Encode the audio */
    i_ret = AUDENC_process( xdm->handle, &xdm->in, &xdm->out,
                            &xdm->in_args, &xdm->out_args );
    if( i_ret != AUDENC_EOK )
    {
        msg_Err( p_enc, "Audio encoding failed (%d): %s", i_ret,
                 davinci_GetExtendedError( xdm->out_args.extendedError ) );
        return NULL;
    }

#if 0
    /* Print some info */
    msg_Dbg( p_enc, "Bytes generated: %d", (int)xdm->out_args.bytesGenerated );
#endif

    /* Put everything in the block */
    if( xdm->out_args.bytesGenerated <= 0 )
        return NULL;

    p_block = block_New( p_enc, xdm->out_args.bytesGenerated );
    if( !p_block )
        return NULL;

    memcpy( p_block->p_buffer, xdm->out.bufs[0], xdm->out_args.bytesGenerated );

    p_block->i_length = i_pts_end - i_pts;
    p_block->i_dts = p_block->i_pts = i_pts;

    return p_block;
}

