/*****************************************************************************
 * davinci.c: decoder and encoder modules using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008-2010 M2X BV
 * $Id: 9edb9c958ad5c665858e905c13e835af1c33b87b $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *          Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_codec.h>

#include <assert.h>

#include "davinci.h"

/*****************************************************************************
 * Module descriptor
 *****************************************************************************/

/*
 * Davinci common
 */

#define DAVINCI_ENGINE_TEXT N_( "DSP firmware library" )
#define DAVINCI_ENGINE_LONGTEXT N_( \
    "Specifiy which DSP firmware library to use. By default " \
    "a combined encoder/decoder library \"encodedecode\" is used." )

/*
 * Decoder
 */

#define DAVINCI_FOURCC_TEXT N_( "Decoder output fourcc/chroma" )
#define DAVINCI_FOURCC_LONGTEXT N_( \
    "Force audio/video decoder to output in specific fourcc/chroma" )

/*
 * Encoder
 */

/* video settings */
#define ENC_KEYINT_TEXT N_( "Ratio of key frames" )
#define ENC_KEYINT_LONGTEXT N_( "Number of frames " \
  "that will be coded for one key frame." )

#define ENC_INTERLACE_TEXT N_( "Interlaced encoding" )
#define ENC_INTERLACE_LONGTEXT N_( "Enable dedicated " \
  "algorithms for interlaced frames." )

#define ENC_RC_TEXT N_( "Rate control Preset method (default: 0)" )
#define ENC_RC_LONGTEXT N_( "Rate control Preset method " \
    "ranges from 1..5 (1=cbr, 2=storage, 3=twopass, 4=none, 5=userdefined)")

#define ENC_VT_TEXT N_( "Bitrate tolerance" )
#define ENC_VT_LONGTEXT N_( "Bitrate tolerance in kbit/s." )

#define ENC_QUALITY_TEXT N_("Encoder quality setting (default: 0)" )
#define ENC_QUALITY_LONGTEXT N_( "Encoder qualtiy setting preset. " \
    "ranges from 0..3 (0=encoders default, 1=high quality, 2=speed, 3=user defined" )

/* XDM 1.0 */
#define ENC_CBR_TEXT N_("Constant bitrate encoding (default: true)" )
#define ENC_CBR_LONGTEXT N_( "Use constant bitrate encoding. If disabled, " \
    "then variable bitrate encoding is used resulting in higher bandwidth savings." )

/* audio settings */
#define ENC_PROFILE_TEXT N_( "Specify AAC audio profile to use" )
#define ENC_PROFILE_LONGTEXT N_( "Specify the AAC audio profile to use " \
   "for encoding the audio bitstream. It takes the following options: " \
   "low (LC), sbr (HE) and ps (HEv2) (default: low)" )

/* */
#ifdef ENABLE_SOUT
static const char *const enc_aac_profile_list[] = { "low", "sbr", "ps" };
static const char *const enc_aac_profile_list_text[] = {
    N_("low"), N_("sbr"), N_("ps") };
#endif

/* */
#define MODULE_DESCRIPTION N_( "Various audio and video decoders/encoders" \
        "delivered by TI Davinci library. This includes MPEG4, H264, VC-1, "\
        "MPEG1, MPEG2, AAC and MP3 codecs")

vlc_module_begin()
    /* audio/video decoder */
    set_shortname( "Davinci" )
    add_shortcut( "davinci" )
    set_category( CAT_INPUT )
    set_subcategory( SUBCAT_INPUT_SCODEC )
    set_description( N_("DaVinci DSP audio/video decoder") )
    set_help( MODULE_DESCRIPTION )
    set_capability( "decoder", 1337 )
    set_section( N_("Decoding") , NULL )
    set_callbacks( OpenVideoDecoder, CloseVideoDecoder )

    add_string( DEC_CFG_PREFIX "engine", "encodedecode", NULL,
                DAVINCI_ENGINE_TEXT, DAVINCI_ENGINE_LONGTEXT, true )

    /* video decoder */
    add_string( DEC_CFG_PREFIX "fourcc", "", NULL,
                DAVINCI_FOURCC_TEXT, DAVINCI_FOURCC_LONGTEXT, true )

    /* audio decoder */
    add_submodule()
    add_shortcut( "davinci" )
    set_description( _("DaVinci DSP audio decoder") );
    set_capability( "decoder", 0 );
    set_callbacks( OpenAudioDecoder, CloseAudioDecoder );

#ifdef ENABLE_SOUT
    /* encoder */
    add_submodule()
    add_shortcut( "davinci" )
    set_description( N_("DaVinci DSP encoder" ) )
    set_capability( "encoder", 1337 )
    set_section( N_("Encoding") , NULL )
    set_callbacks( OpenEncoder, CloseEncoder )

    /* encoder settings - XDM 0.9 only */
    add_integer( ENC_CFG_PREFIX "quality", 0, NULL, ENC_QUALITY_TEXT,
                 ENC_QUALITY_LONGTEXT, false )
        change_integer_range( 0, 3 )

    /* encoder settings - XDM 1.0 only */
    add_bool( ENC_CFG_PREFIX "cbr", true, NULL, ENC_CBR_TEXT,
              ENC_CBR_LONGTEXT, false )

    /* video encoder settings - XDM all versions */
    add_integer( ENC_CFG_PREFIX "keyint", 0, NULL, ENC_KEYINT_TEXT,
                 ENC_KEYINT_LONGTEXT, false )
    add_bool( ENC_CFG_PREFIX "interlace", false, NULL, ENC_INTERLACE_TEXT,
              ENC_INTERLACE_LONGTEXT, false )
    add_integer( ENC_CFG_PREFIX "rc", 1, NULL, ENC_RC_TEXT,
                 ENC_RC_LONGTEXT, false )
        change_integer_range( 0, 5 )
    add_integer( ENC_CFG_PREFIX "tolerance", 0, NULL, ENC_VT_TEXT,
                 ENC_VT_LONGTEXT, false )

    /* audio encoder settings */
    add_string( ENC_CFG_PREFIX "aac-profile", "low", NULL,
                ENC_PROFILE_TEXT, ENC_PROFILE_LONGTEXT, false )
        change_string_list( enc_aac_profile_list, enc_aac_profile_list_text, 0 )

#endif
#if 0
    /* video filter submodule */
    add_submodule ()
    set_capability( "video filter2", 0 )
    set_callbacks( OpenResizer, CloseResizer )
    set_description( N_("Davinci resizer video filter") )
    add_shortcut( "davinci-resizer" )
#endif
vlc_module_end()

/*****************************************************************************
 * Memory utils
 *****************************************************************************/
#include <ti/sdo/ce/osal/Memory.h>

/* XDM 0.9 */
int davinci_AllocateBuffer( XDAS_Int32 i_num, XDAS_Int32 *pi_sizes,
                            XDM_BufDesc *buf )
{
    int i;

    buf->numBufs = i_num;

    buf->bufs = (XDAS_Int8 **)malloc( buf->numBufs * sizeof( XDAS_Int8 * ) );
    if( !buf->bufs )
    {
        buf->numBufs = 0;
        return VLC_ENOMEM;
    }

    buf->bufSizes = (XDAS_Int32 *)malloc( buf->numBufs * sizeof( XDAS_Int32 ) );
    if( !buf->bufSizes )
    {
        free( buf->bufs );
        buf->numBufs = 0;
        return VLC_ENOMEM;
    }

    for( i = 0; i < i_num; i++ )
    {
        buf->bufSizes[i] = pi_sizes[i];
        buf->bufs[i] = Memory_contigAlloc( pi_sizes[i], Memory_DEFAULTALIGNMENT );
        if( !buf->bufs[i] )
        {
            for( i--; i >= 0; i-- )
                Memory_contigFree( buf->bufs[i], buf->bufSizes[i] );

            free( buf->bufs );
            free( buf->bufSizes );
            buf->numBufs = 0;
            return VLC_ENOMEM;
        }
    }
    return VLC_SUCCESS;
}

void davinci_FreeBuffer( XDM_BufDesc *buf )
{
    if( buf->numBufs != 0 )
    {
        int i;
        for( i = 0; i < buf->numBufs; i++ )
            Memory_contigFree( buf->bufs[i], buf->bufSizes[i] );
        free( buf->bufs );
        free( buf->bufSizes );
        buf->numBufs = 0;
    }
}

/* XDM 1.0 */
int davinci_AllocateBuffer1( XDAS_Int32 i_num, XDAS_Int32 *pi_sizes,
                             XDM1_BufDesc *buf )
{
    int i;

    assert(i_num < XDM_MAX_IO_BUFFERS);

    buf->numBufs = i_num;
    for( i = 0; i < i_num; i++ )
    {
        buf->descs[i].bufSize = pi_sizes[i];
        buf->descs[i].buf = Memory_contigAlloc( pi_sizes[i], Memory_DEFAULTALIGNMENT );
        if( !buf->descs[i].buf )
        {
            goto error;
        }
    }
    return VLC_SUCCESS;

error:
    for( i--; i >= 0; i-- )
         Memory_contigFree( buf->descs[i].buf, buf->descs[i].bufSize );

    return VLC_ENOMEM;
}

void davinci_FreeBuffer1( XDM1_BufDesc *buf )
{
    if( buf->numBufs != 0 )
    {
        int i;
        for( i = 0; i < buf->numBufs; i++ )
            Memory_contigFree( buf->descs[i].buf, buf->descs[i].bufSize );
        buf->numBufs = 0;
    }
}

/*****************************************************************************
 * Misc utils
 *****************************************************************************/
const char *ppsz_engine_error[] = {
    [Engine_EOK] = "Ok",
    [Engine_EEXIST] = "Engine name doesn't exist",
    [Engine_ENOMEM] = "Can't allocate engine memory",
    [Engine_EDSPLOAD] = "Unable to load the DSP image, did you try to open 2 different engines because you can only have one on the DSP, but you can access it with multiple threads.",
    [Engine_ENOCOMM] = "Can't create communication connection to DSP",
    [Engine_ENOSERVER] = "Can't locate the server on the DSP",
    [Engine_ECOMALLOC] = "Can't allocate communication buffer",
    [Engine_ERUNTIME] = "Engine runtime failure",
    [Engine_ECODECCREATE] = "Engine codec creation failed",
    [Engine_ECODECSTART] = "Engine codec start failed",
    [Engine_EINVAL] = "Bad parameter",
    [Engine_EBADSERVER] = "Incompatible server specified",
    [Engine_ENOTAVAIL] = "Service not available",
    //[Engine_EWRONGSTATE] = "Call can't be made at this time",
    //[Engine_EINUSE] = "Call can't be made at this time because a required resource is in use",
    //[Engine_ENOTFOUND] = "Entity was not found",
};

const char * davinci_GetExtendedError( XDAS_Int32 error )
{
    if( XDM_ISAPPLIEDCONCEALMENT( error ) )
        return "Applied bit concealment";
    if( XDM_ISINSUFFICIENTDATA( error ) )
        return "Insufficient data";
    if( XDM_ISCORRUPTEDDATA( error ) )
        return "Corrupted data";
    if( XDM_ISCORRUPTEDHEADER( error ) )
        return "Corrupted header";
    if( XDM_ISUNSUPPORTEDINPUT( error ) )
        return "Unsupported input";
    if( XDM_ISUNSUPPORTEDPARAM( error ) )
        return "Unsupported param";
    if( XDM_ISFATALERROR( error ) )
        return "Fatal error";

    return "Unkown error";
}

void davinci_PrintAvailableAlgorithms( vlc_object_t *p_this, const char *psz_engine )
{
    int i = 0;
    Engine_AlgInfo info;
    info.algInfoSize = sizeof( info );
    msg_Dbg( p_this, "Available algorithms in engine `%s':", psz_engine );
    while( Engine_getAlgInfo( (String)psz_engine, &info, i ) == Engine_EOK )
    {
        const char **psz_type = (const char **)info.typeTab;
        i++;
        msg_Dbg( p_this, "    %d: %s (%s)", i, info.name, info.isLocal?"local":"remote" );
        if( *psz_type )
        {
            msg_Dbg( p_this, "        Inheritance hierarchy:" );
            for( ; *psz_type; psz_type++ )
                msg_Dbg( p_this, "            %s", *psz_type );
        }
    }
}

/* NOTE:
 * CERuntime should be inited and exited only once per application.
 */
static void vlc_ceruntime_mutex (bool acquire)
{
    static vlc_mutex_t lock = VLC_STATIC_MUTEX;

    if (acquire)
        vlc_mutex_lock (&lock);
    else
        vlc_mutex_unlock (&lock);
}

static inline void vlc_ceruntime_lock()
{
    vlc_ceruntime_mutex(true);
}

static inline void vlc_ceruntime_unlock()
{
    vlc_ceruntime_mutex(false);
}

static bool b_ceruntimeinit = false;
static uint32_t ref_ceruntime = 0;

void davinci_RuntimeInit( vlc_object_t *p_this )
{
    vlc_ceruntime_lock();

    ref_ceruntime++;
    if( !b_ceruntimeinit )
    {
        /* Initialize the codec engine */
        CERuntime_init();
        b_ceruntimeinit = true;

        msg_Dbg( p_this, "CERuntime initialized" );
    }
    else
    {
        msg_Dbg( p_this, "CERuntime already initialized" );

    }
    vlc_ceruntime_unlock();
}

void davinci_RuntimeExit( vlc_object_t *p_this )
{
    vlc_ceruntime_lock();

    ref_ceruntime--;
    if( ref_ceruntime == 0 )
    {
        /* Exit the codec engine */
        CERuntime_exit();
        b_ceruntimeinit = false;
        msg_Dbg( p_this, "CERuntime exited" );
    }
    else
    {
        msg_Dbg( p_this, "CERuntime exited (reference count %d)", ref_ceruntime );
    }
    vlc_ceruntime_unlock();
}

