/*****************************************************************************
 * encoder.c: video encoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008-2010 M2X BV
 * $Id: 5daccda5bfdbc37e406fe2c1fc27bdab54391658 $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *          Rafaël Carré <rcarre@m2x.nl>
 *          Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_vout.h>
#include <vlc_aout.h>
#include <vlc_sout.h>
#include <vlc_codec.h>

#undef DEBUG_FPS
#include "davinci.h"

#include <assert.h>

/*****************************************************************************
 * Encoder
 *****************************************************************************/
static const char *const ppsz_enc_options[] = {
    "aac-profile", "cbr", "keyint", "interlace", "quality",
    "rc", "tolerance",
    NULL
};

/*****************************************************************************
 *
 *****************************************************************************/
int OpenEncoder( vlc_object_t *p_this )
{
    encoder_t *p_enc = (encoder_t *)p_this;
    encoder_sys_t *p_sys = NULL;
    Engine_Error err;
    int i_cat, i_xdm;
    const char *psz_namecodec;

    /* Allocate our private structure */
    p_enc->p_sys = (encoder_sys_t *)calloc( 1, sizeof( encoder_sys_t ) );
    if( !p_enc->p_sys )
        return VLC_ENOMEM;
    p_sys = p_enc->p_sys;

    if( !GetDavinciEncoder( p_enc->fmt_out.i_codec, &i_cat, &i_xdm,
                            &p_sys->psz_codec, &psz_namecodec ) )
    {
        msg_Dbg( p_enc, "Unsupported codec : %4.4s",(char*)&p_enc->fmt_out.i_codec );
        free( p_sys );
        return VLC_EGENERIC;
    }

    msg_Dbg( p_enc, "found davinci encoder %s for %s (codec=%4.4s, cat=%s)",
                p_sys->psz_codec, (char*)&p_enc->fmt_out.i_codec, psz_namecodec,
                (i_cat == VIDEO_ES) ? "video" :
                    (i_cat == AUDIO_ES) ? "audio" : "other" );

    p_sys->psz_ti_engine = var_CreateGetString( p_enc, DEC_CFG_PREFIX "engine" );
    if( !p_sys->psz_ti_engine )
    {
        free( p_sys );
        return VLC_ENOMEM;
    }

    /* Initialize the codec engine */
    davinci_RuntimeInit( p_this );

    /* Create an engine handle */
    p_sys->engine = Engine_open( p_sys->psz_ti_engine, NULL /*&Engine_ATTRS*/, &err );
    if( err != Engine_EOK )
    {
        msg_Err( p_enc, "Error while opening engine `%s': %s",
                 p_sys->psz_ti_engine, ppsz_engine_error[err] );
        goto error;
    }

    davinci_PrintAvailableAlgorithms( p_this, p_sys->psz_ti_engine );

    config_ChainParse( p_enc, ENC_CFG_PREFIX, ppsz_enc_options, p_enc->p_cfg );

    /* Initialize random stuff */
    switch( i_cat )
    {
        case VIDEO_ES:
            if( OpenEncoderVideo( p_enc ) != VLC_SUCCESS )
                goto error;
            p_enc->pf_encode_video = p_sys->pf_encode_video;
            break;
        case AUDIO_ES:
            switch( i_xdm )
            {
                case DAVINCI_XDM0_9:
                    if( OpenEncoderAudio( p_enc ) != VLC_SUCCESS )
                        goto error;
                    break;
                case DAVINCI_XDM1_0:
                    if( OpenEncoderAudio1( p_enc ) != VLC_SUCCESS )
                        goto error;
                    break;
                default:
                    goto error;
            }
            p_enc->pf_encode_audio = p_sys->pf_encode_audio;
            break;
        default:
            msg_Err( p_enc, "aborting unknown ES (%d)", i_cat );
            goto error;
    }

    return VLC_SUCCESS;

error:
    msg_Err( p_enc, "Failed to create %s encoder (%s) for %s",
                (i_cat == VIDEO_ES) ? "video" :
                    (i_cat == AUDIO_ES) ? "audio" : "other",
                p_sys->psz_codec, psz_namecodec );

    CloseEncoder( p_this );
    return VLC_EGENERIC;
}

/*****************************************************************************
 * CloseEncoder
 *****************************************************************************/
void CloseEncoder( vlc_object_t *p_this )
{
    encoder_t *p_enc = (encoder_t *)p_this;
    encoder_sys_t *p_sys = p_enc->p_sys;

    /* Delete encoder */
    int i_xdm, i_cat;
    if( !GetDavinciEncoder( p_enc->fmt_out.i_codec, &i_cat, &i_xdm,
                            NULL, NULL ) )
    {
        switch( i_cat )
        {
        case VIDEO_ES:
            switch( i_xdm )
            {
                case DAVINCI_XDM0_9: CloseEncoderVideo( p_enc ); break;
                //case DAVINCI_XDM1_0: CloseEncoderVideo1( p_enc ); break;
            }
            break;
        case AUDIO_ES:
            switch( i_xdm )
            {
                case DAVINCI_XDM0_9: CloseEncoderAudio( p_enc ); break;
                case DAVINCI_XDM1_0: CloseEncoderAudio1( p_enc ); break;
            }
            break;
        default:
            msg_Err( p_enc, "Trying to release unsupported codec." );
            break;
        }
    }

    /* Delete codec engine handle */
    Engine_close( p_sys->engine );

    /* Exit the codec engine */
    davinci_RuntimeExit( p_this );

    free( p_sys->psz_ti_engine );
    free( p_sys );
}

#ifdef DEBUG_FPS
void calculate_fps( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;

    p_sys->i_pics++;
    if( p_sys->i_pics == 1 )
    {
        p_sys->i_pts_start = mdate();
    }

    if( p_sys->i_pics == p_sys->i_fps )
    {
        mtime_t i_diff; /* in us */
        double f_fps;

        p_sys->i_pts_end = mdate(); /* in us */
        i_diff = (p_sys->i_pts_end - p_sys->i_pts_start);
        f_fps = (double)((i_diff / (int64_t)p_sys->i_pics)); /* 1 frame per us */
        f_fps = ((double)1/f_fps) * (double)1000000;
        msg_Info( p_enc, "%d frames per %"PRId64" micro seconds (%4.2f fps)",
                  p_sys->i_pics, i_diff, f_fps );

        /* reset timers */
        p_sys->i_pics = 0;
        p_sys->i_pts_start = p_sys->i_pts_end = (mtime_t)0;
    }
}
#endif

