/*****************************************************************************
 * resizer.c : DaVinci resizer
 *****************************************************************************
 * Copyright (C) 2008-2010 M2X BV
 * Copyright © 2010 Rafaël Carré
 * $Id: 974bf03cdb7d90f30d7f83f97e24bd870c202799 $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#undef _FILE_OFFSET_BITS        /* mmap() will fail if we use 64bits offsets */

#include <vlc_common.h>

#include <stdio.h>
#include <string.h>                                            /* strerror() */
#include <stdlib.h>                                      /* malloc(), free() */
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include <sys/mman.h>
#include <sys/ioctl.h>

#include <gnu/targets/std.h>
#include <xdc/std.h>
#include <ti/sdo/ce/osal/Memory.h>
#include <davinci_resizer.h>

#include "resizer.h"

/*****************************************************************************
 *
 *****************************************************************************/
typedef struct resizer_t
{
    int fd; /* /dev/davincresizer */

    rsz_resize_t resize;

    int in_size;
    uint8_t *p_out;
    int out_size;

    int in_height;
    int in_width;
    int out_height;
    int out_width;
} davinci_resizer_t;

/* resizer_coefs.c */
extern void resize_coefs(unsigned int vidw, unsigned int vidh,
    unsigned int *width, unsigned int *height,
    short hcoefs[32], short vcoefs[32]);

/*****************************************************************************
 * Local functions
 *****************************************************************************/
static uint8_t *GetBuffer(struct resizer_t *rsz, int type, int size, int *offset)
{
    uint8_t *p_buf;
    struct rsz_reqbufs reqbufs;
    struct rsz_buffer buffer;

    memset(&reqbufs, 0, sizeof(reqbufs));
    memset(&buffer, 0, sizeof(buffer));

    reqbufs.buf_type = type;
    reqbufs.size = size;
    reqbufs.count = 1;

    if (ioctl(rsz->fd, RSZ_REQBUF, &reqbufs))
        return NULL;

    buffer.buf_type = type;
    buffer.index = 0;

    if (ioctl(rsz->fd, RSZ_QUERYBUF, &buffer))
        return NULL;

    if (offset) *offset = buffer.offset;

    p_buf = mmap(NULL, reqbufs.size, PROT_READ|PROT_WRITE, MAP_SHARED,
                  rsz->fd, buffer.offset);
    if(p_buf == MAP_FAILED)
        return NULL;

    return p_buf;
}

static int Init(vlc_object_t *p_this, struct resizer_t *rsz, int in_width, int in_height,
        int *out_width, int *out_height)
{
    /* Set the parameters */
    rsz_params_t params;
    memset(&params, 0, sizeof(params));
    params.in_hsize = in_width;
    params.in_vsize = in_height;
    params.in_pitch = OUT_WIDTH << 1;
    params.inptyp = RSZ_INTYPE_YCBCR422_16BIT;
    params.pix_fmt = RSZ_PIX_FMT_YUYV;

    /* will modify output dimensions to something sensible */
    resize_coefs(params.in_hsize, params.in_vsize, out_width, out_height,
                 params.hfilt_coeffs, params.vfilt_coeffs);

    params.out_hsize = *out_width;
    params.out_vsize = *out_height;
    params.out_pitch = OUT_WIDTH << 1;

    params.yenh_params.type = RSZ_YENH_DISABLE;

    if (ioctl(rsz->fd, RSZ_S_PARAM, &params))
    {
        msg_Err(p_this, "Failed setting resizer params (look at dmesg)\n");
        return -1;
    }

    /* Set resizer speed to fastest */
    int speed = 0;
    if (ioctl(rsz->fd, RSZ_S_EXP, &speed))
    {
        msg_Err(p_this, "Failed setting resizer speed\n");
        return -2;
    }

    rsz->in_size = (OUT_WIDTH << 1) * in_height;
    rsz->resize.in_buf.index = -1;

    rsz->out_size = (OUT_WIDTH << 1) * *out_height;
#ifdef DIRECT_OUTPUT
    rsz->resize.out_buf.index = -1;
#else
    rsz->resize.out_buf.index = 0;
    rsz->p_out = GetBuffer(rsz, RSZ_BUF_OUT, rsz->out_size,
            &rsz->resize.out_buf.offset); /* FIXME: offset width + height */
    if (!rsz->p_out)
        return -4;
#endif

    /* Initialize resize structure */
    rsz->resize.in_buf.size = rsz->in_size;
    rsz->resize.out_buf.size = rsz->out_size;

    return 0;
}

/*****************************************************************************
 * Exported functions
 *****************************************************************************/
void resizer_close(struct resizer_t *rsz)
{
    close(rsz->fd);
    free(rsz);
}

struct resizer_t *resizer_open(vlc_object_t *p_this,
    int in_width, int in_height, int *out_width, int *out_height)
{
    struct resizer_t *rsz = calloc(1, sizeof(*rsz));
    if (!rsz)
        return NULL;

    rsz->fd = open("/dev/davinci_resizer", O_RDWR);
    if (rsz->fd == -1)
    {
        msg_Err(p_this, "Couldn't open device (%m)");
        free(rsz);
        return NULL;
    }

    int ret = Init(p_this, rsz, in_width, in_height, out_width, out_height);
    if (ret != 0)
    {
        msg_Err(p_this, "Resizer init failed (%d)\n", ret);
        resizer_close(rsz);
        return NULL;
    }

    rsz->in_height = in_height;
    rsz->in_width = in_width;

    rsz->out_height = *out_height;
    rsz->out_width = *out_width;

    return rsz;
}

/*  fill pointer with size bytes of black (== size/2 pixels!)
 *  and increment pointer */
static inline uint8_t *fill_black(uint8_t *out, const size_t size)
{
    assert((size & 3) == 0);
    if(!size)
        return out;

    register int loops = size >> 2;
    register const int black = 0x10801080;

#if 1   /* ASM */
    asm volatile (
      "1:\tstr %[b], [%[out]], #4   \n" /* do { *p++ = black;         */
        "\tsubs %[s], %[s], #1      \n" /* } while(--size);           */
        "\tbne 1b                   \n"

/* output  */ : [out]"+r"(out), [s]"+r"(loops)  /* both are modified */
/* input   */ : [b]"r"(black)
/* clobber */ : "memory"
    );

#else   /* C */
    int *out32 = (int*) out;
    do {
        *out32++ = black;
    } while(--loops);
    out = (uint8_t*)out32;
#endif  /* C / ASM */

    return out;
}

int resizer_resize(struct resizer_t *rsz, uint8_t *in, uint8_t *out)
{
    /* We can use the decoder output buffer as is */
    rsz->resize.in_buf.offset =
            Memory_getBufferPhysicalAddress(in, rsz->in_size, NULL);

#ifdef DIRECT_OUTPUT
    rsz->resize.out_buf.offset =
        Memory_getBufferPhysicalAddress(out, rsz->out_size, NULL);
    assert(rsz->resize.out_buf.offset);

    const unsigned lines = (OUT_HEIGHT - rsz->out_height);
    const unsigned pitch = OUT_WIDTH << 1;
    const unsigned cols = (OUT_WIDTH - rsz->out_width); /* in bytes */

    rsz->resize.out_buf.offset += (lines >> 1) * pitch + (cols & ~31);
#endif

    int error;
    do
    {   /* resize */
        error = ioctl(rsz->fd, RSZ_RESIZE, &rsz->resize);
    } while(error == -1 && errno == EAGAIN);

    if (error == -1)
        return errno;

#ifdef DIRECT_OUTPUT
    const unsigned cols_left = (cols & ~31);
    const unsigned cols_right = ((OUT_WIDTH - rsz->out_width) << 1) - cols_left;
    const unsigned lines_top = lines >> 1;
    const unsigned lines_bottom = (lines + 1) >> 1;
    const unsigned pitch_pic = rsz->out_width << 1;

    assert((pitch_pic & 3) == 0);

    out = fill_black(out, lines_top * pitch);
    for (int i = 0; i < rsz->out_height; i++)
    {
        out = fill_black(out, cols_left) + pitch_pic;
        out = fill_black(out, cols_right);
    }
    fill_black(out, lines_bottom * pitch);
#else
    resizer_copy_picture(out, rsz->p_out, rsz->out_width, rsz->out_height);
#endif

    return 0;
}

/*
 *  copy input  picture (width     x height)
 *  into output picture (OUT_WIDTH x OUT_HEIGHT)
 *
 *  fill borders with black
 */
void resizer_copy_picture(uint8_t *out, uint8_t *in, int width, int height)
{
    const unsigned h_offset = OUT_HEIGHT - height;
    const unsigned h_black_top =    (OUT_WIDTH << 1) *  (h_offset      >> 1);
    const unsigned h_black_bottom = (OUT_WIDTH << 1) * ((h_offset + 1) >> 1);

    const unsigned pitch = (width << 1) & ~3;

    const unsigned v_black_left = (OUT_WIDTH - width) & ~3;
    const unsigned v_black_right = (OUT_WIDTH << 1) - pitch - v_black_left;

    out = fill_black(out, h_black_top);

    for (int i = 0; i < height; i++)
    {
        out = fill_black(out, v_black_left);

        memcpy(out, in, pitch);
        in  += (OUT_WIDTH << 1);
        out += pitch;

        out = fill_black(out, v_black_right);
    }

    fill_black(out, h_black_bottom);
}
