/*****************************************************************************
 * video.c: video decoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008-2010 M2X BV
 * $Id: 4e587f0da14bf19df43af7c58b2b285ad762fca2 $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *          Rafaël Carré <rcarre@m2x.nl>
 *          Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_codec.h>
#include <vlc_vout.h>
#include <vlc_codecs.h>

#include "davinci.h"

#include <ti/sdo/ce/video/viddec.h>

#include <assert.h>

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static picture_t *DecodeVideoBlock( decoder_t *, block_t ** );

/*****************************************************************************
 * Video decoder
 *****************************************************************************/
struct decoder_sys_t
{
    char *psz_ti_engine;
    Engine_Handle engine;
    VIDDEC_Handle handle;

    VIDDEC_InArgs in_args;
    VIDDEC_OutArgs out_args;

    VIDDEC_DynamicParams dparams;
    VIDDEC_Status status;

    XDM_BufDesc in;
    XDM_BufDesc out;

    /* picture properties */
    bool b_progressive;
};

/*****************************************************************************
 * OpenDecoder: probe the decoder and return score
 *****************************************************************************/
int OpenVideoDecoder( vlc_object_t *p_this )
{
    decoder_t *p_dec = (decoder_t*)p_this;
    decoder_sys_t *p_sys;
    Engine_Error err;
    int i_cat, i_xdm;
    char *psz_chroma = NULL;
    const char *psz_codec = NULL;
    const char *psz_namecodec = NULL;
    vlc_fourcc_t i_chroma = VLC_FOURCC('U','Y','V','Y');
    VIDDEC_Params params;

    if( !GetDavinciDecoder( p_dec->fmt_in.i_codec, &i_cat, &i_xdm,
                            &psz_codec, &psz_namecodec ) )
        return VLC_EGENERIC;

    if( i_cat != VIDEO_ES ) return VLC_EGENERIC;

    msg_Dbg( p_dec, "found davinci decoder %s for %s (codec=%4.4s, cat=%s)",
                psz_codec, (char*)&p_dec->fmt_out.i_codec, psz_namecodec,
                (i_cat == VIDEO_ES) ? "video" :
                    (i_cat == AUDIO_ES) ? "audio" : "other" );

    /* Allocate our private structure */
    p_dec->p_sys = (decoder_sys_t *)calloc( 1, sizeof( decoder_sys_t ) );
    if( !p_dec->p_sys )
        return VLC_ENOMEM;

    p_sys = p_dec->p_sys;

    /* setup in_args and out_args */
    p_sys->in_args.size = sizeof( p_sys->in_args );
    p_sys->out_args.size = sizeof( p_sys->out_args );
    p_sys->dparams.size = sizeof( VIDDEC_DynamicParams );
    p_sys->status.size = sizeof( VIDDEC_Status );

    p_sys->psz_ti_engine = var_CreateGetString( p_dec, DEC_CFG_PREFIX "engine" );
    if( !p_sys->psz_ti_engine )
    {
        free( p_sys );
        return VLC_ENOMEM;
    }

    /* Initialize the codec engine */
    davinci_RuntimeInit( p_this );

    /* Create an engine handle */
    p_sys->engine = Engine_open( p_sys->psz_ti_engine, NULL /*&Engine_ATTRS*/, &err );
    if( err != Engine_EOK )
    {
        msg_Err( p_dec, "Error while opening engine `%s': %s",
                 p_sys->psz_ti_engine, ppsz_engine_error[err] );
        goto error;
    }

    davinci_PrintAvailableAlgorithms( p_this, p_sys->psz_ti_engine );

    /* Get user supplied chroma setting */
    psz_chroma = var_CreateGetString( p_dec, DEC_CFG_PREFIX "fourcc" );
    if( psz_chroma && strlen( psz_chroma ) == 4 )
    {
        i_chroma = VLC_FOURCC( psz_chroma[0], psz_chroma[1],
                               psz_chroma[2], psz_chroma[3] );
        msg_Dbg( p_dec, "Forcing output chroma to `%s'", psz_chroma );
    }
    free( psz_chroma );

    /* Create video decoder */
    params.size = sizeof( VIDDEC_Params );

    /* TODO change the params when HD codecs are available */
    params.maxHeight = p_dec->fmt_in.video.i_height ?: 576 /*720*/;
    params.maxWidth = p_dec->fmt_in.video.i_width ?: 720 /*1280*/;
    /* The application should set the parameters to 0 to use the algorithm's default values. */
    params.maxFrameRate = 0;
    params.maxBitRate = 0;

    msg_Dbg( p_dec, "Creating decoder for %dx%d codec %4.4s in chroma %4.4s",
        (int)params.maxWidth, (int)params.maxHeight,
        (const char *)&p_dec->fmt_in.i_codec,
        (char*)&i_chroma
        );

    params.dataEndianness = XDM_BYTE;
    params.forceChromaFormat = VlcChromaToXdm( i_chroma );
    if( params.forceChromaFormat == XDM_CHROMA_NA )
    {
        msg_Err( p_dec, "Unsupported output chroma (%4.4s).",
                 (const char *)&i_chroma );
        goto error;
    }

    p_sys->handle = VIDDEC_create( p_sys->engine, (String)psz_codec, &params );
    if( !p_sys->handle )
    {
        msg_Err( p_dec, "Failed to create video decoder (%s) for %s",
                 psz_codec, psz_namecodec );
        goto error;
    }

    /* Set output properties */
    p_dec->fmt_out.i_cat = VIDEO_ES;
    p_dec->fmt_out.i_codec =
        p_dec->fmt_out.video.i_chroma = i_chroma;
    p_dec->fmt_out.video.i_width = p_dec->fmt_in.video.i_width;
    p_dec->fmt_out.video.i_height = p_dec->fmt_in.video.i_height;
    p_dec->fmt_out.video.i_aspect = p_dec->fmt_in.video.i_aspect;

    /* Set callbacks */
    p_dec->pf_decode_video = DecodeVideoBlock;

    /* DaVinci decoder needs complete frames */
    p_dec->b_need_packetized = true;
    return VLC_SUCCESS;

error:
    if( p_sys->engine ) Engine_close( p_sys->engine );
    free( p_sys->psz_ti_engine );
    free( p_sys );
    return VLC_EGENERIC;
}

/*****************************************************************************
 * CloseDecoder: png decoder destruction
 *****************************************************************************/
void CloseVideoDecoder( vlc_object_t *p_this )
{
    decoder_t *p_dec = (decoder_t *)p_this;
    decoder_sys_t *p_sys = p_dec->p_sys;

    /* Close our codec handle */
    VIDDEC_delete( p_sys->handle );

    /* Close our engine handle */
    Engine_close( p_sys->engine );

    /* Exit the codec engine */
    davinci_RuntimeExit( p_this );

    /* Free 'DaVinci compliant' buffers */
    davinci_FreeBuffer( &p_sys->in );
    davinci_FreeBuffer( &p_sys->out );

    free( p_sys->psz_ti_engine );
    free( p_sys );
}

/* */
static inline void davinci_CopyXDMtoPicture( picture_t *p_pic, XDM_BufDesc *p_buf )
{
    /* Copy stuff to picture here */
    assert( p_pic->i_planes == p_buf->numBufs );

    for( int i = 0; i < p_pic->i_planes; i++ )
    {
        plane_t *p = p_pic->p + i;
        memcpy( p->p_pixels, p_buf->bufs[i], p->i_pitch * p->i_visible_lines );
    }
}

#ifdef DAVINCI_DEBUG
static void davinci_FrameType( decoder_t *p_dec, IVIDEO_FrameType frametype )
{
    const char *frame;
    switch( frametype )
    {
    case IVIDEO_NA_FRAME: frame = "unknown"; break;
    case IVIDEO_I_FRAME:  frame = "I"; break;
    case IVIDEO_P_FRAME:  frame = "P"; break;
    case IVIDEO_B_FRAME:  frame = "B"; break;
    case IVIDEO_IDR_FRAME:frame = "IDR"; break;
    case IVIDEO_II_FRAME: frame = "I interlaced"; break;
    case IVIDEO_IP_FRAME: frame = "IP interlaced"; break;
    case IVIDEO_IB_FRAME: frame = "IB interlaced"; break;
    case IVIDEO_PI_FRAME: frame = "PI interlaced"; break;
    case IVIDEO_PP_FRAME: frame = "PP interlaced"; break;
    case IVIDEO_PB_FRAME: frame = "PB interlaced"; break;
    case IVIDEO_BI_FRAME: frame = "BI interlaced"; break;
    case IVIDEO_BP_FRAME: frame = "BP interlaced"; break;
    case IVIDEO_BB_FRAME: frame = "BB interlaced"; break;
    case IVIDEO_MBAFF_I_FRAME: frame = "MBAFF I"; break;
    case IVIDEO_MBAFF_P_FRAME: frame = "MBAFF P"; break;
    case IVIDEO_MBAFF_B_FRAME: frame = "MBAFF B"; break;
    case IVIDEO_MBAFF_IDR_FRAME: frame = "MBAFF IDR"; break;
    default: frame = "other"; break;
    }
    msg_Dbg( p_dec, "Frametype: %s", frame );
}
#endif

/****************************************************************************
 * DecodeBlock: the whole thing
 ****************************************************************************/
static picture_t *DecodeVideoBlock( decoder_t *p_dec, block_t **pp_block )
{
    decoder_sys_t *p_sys = p_dec->p_sys;
    block_t *p_block;
    picture_t *p_pic = NULL;

    if( !pp_block || !*pp_block ) return NULL;
    p_block = *pp_block;

    if( p_block->i_flags & (BLOCK_FLAG_DISCONTINUITY|BLOCK_FLAG_CORRUPTED) )
    {
        block_Release( p_block );
        return NULL;
    }

    /* Configure video decoder */
    if( p_sys->in.numBufs == 0 || p_sys->out.numBufs == 0 )
    {
        if( VIDDEC_control( p_sys->handle, XDM_GETBUFINFO, &p_sys->dparams, &p_sys->status )
            != VIDDEC_EOK )
        {
            msg_Err( p_dec, "Failed to get buffer info" );
            goto error;
        }
    }
    if( p_sys->in.numBufs == 0 )
    {
        /* Allocate input buffer */
        if( davinci_AllocateBuffer( p_sys->status.bufInfo.minNumInBufs,
                                    p_sys->status.bufInfo.minInBufSize, &p_sys->in )
            != VLC_SUCCESS )
        {
            msg_Err( p_dec, "Failed to allocate input buffers" );
            goto error;
        }
    }
    if( p_sys->out.numBufs == 0 )
    {
        /* Allocate output buffer(s) */
        if( davinci_AllocateBuffer( p_sys->status.bufInfo.minNumOutBufs,
                                    p_sys->status.bufInfo.minOutBufSize, &p_sys->out )
            != VLC_SUCCESS )
        {
            davinci_FreeBuffer( &p_sys->in );
            msg_Err( p_dec, "Failed to allocate output buffers" );
            goto error;
        }
    }

#ifdef DAVINCI_DEBUG
    if( VIDDEC_control( p_sys->handle, XDM_GETSTATUS, &p_sys->dparams, &p_sys->status )
        != VIDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to get decoder status" );
        goto error;
    }
#endif

    /* Setup input buffer */
    p_sys->in_args.numBytes = __MIN( p_dec->fmt_in.i_extra + p_block->i_buffer,
                                     (size_t)p_sys->in.bufSizes[0] );
    p_sys->in_args.inputID++;

    assert( p_block->i_buffer <= (size_t)p_sys->in.bufSizes[0] );

    if( p_dec->fmt_in.i_extra > 0 )
    {
        memcpy( p_sys->in.bufs[0], p_dec->fmt_in.p_extra, p_dec->fmt_in.i_extra );
    }
    memcpy( p_sys->in.bufs[0] + p_dec->fmt_in.i_extra, p_block->p_buffer,
            p_sys->in_args.numBytes - p_dec->fmt_in.i_extra );

    /* Decode the video */
    if( VIDDEC_process( p_sys->handle, &p_sys->in, &p_sys->out,
                        &p_sys->in_args, &p_sys->out_args ) != VIDDEC_EOK )
    {
        msg_Err( p_dec, "Video decoding failed: %s",
            davinci_GetExtendedError( p_sys->out_args.extendedError ) );
        goto error;
    }

    if( p_dec->fmt_out.video.i_width == 0 ||
        p_dec->fmt_out.video.i_height == 0 ||
        p_dec->fmt_out.video.i_aspect == 0 )
    {
        if( VIDDEC_control( p_sys->handle, XDM_GETSTATUS,
                            &p_sys->dparams, &p_sys->status ) != VIDDEC_EOK )
        {
            msg_Err( p_dec, "Failed to get decoder status" );
            goto error;
        }

        p_dec->fmt_out.video.i_width = p_sys->status.outputWidth;
        p_dec->fmt_out.video.i_height = p_sys->status.outputHeight;
        p_dec->fmt_out.video.i_aspect = VOUT_ASPECT_FACTOR * p_sys->status.outputWidth
                                        / p_sys->status.outputHeight; /* FIXME */

        p_sys->b_progressive = (p_sys->status.contentType == IVIDEO_INTERLACED) ? false : true;

#ifdef DEBUG_DAVINCI
        msg_Dbg( p_dec, "Frame is %d bytes (%d extra), %dx%d, type %s",
                (int)p_sys->in_args.numBytes, p_dec->fmt_in.i_extra,
                (int)p_sys->status.outputWidth, (int)p_sys->status.outputHeight,
                  p_sys->status.contentType == IVIDEO_PROGRESSIVE ? "progressive" :
                  p_sys->status.contentType == IVIDEO_INTERLACED ? "interlaced" :
                  "unknown" );
#endif
    }

#ifdef DEBUG_DAVINCI
    if( p_sys->in_args.numBytes > p_sys->out_args.bytesConsumed )
        msg_Err( p_dec, "%d bytes were not consumed",
                 (int)(p_sys->out_args.bytesConsumed - p_sys->in_args.numBytes) );

    davinci_FrameType( p_dec, p_sys->out_args.decodedFrameType );
#endif

    p_block->p_buffer += p_sys->out_args.bytesConsumed - p_dec->fmt_in.i_extra;
    p_block->i_buffer -= p_sys->out_args.bytesConsumed - p_dec->fmt_in.i_extra;

    /* Get a new picture */
    p_pic = decoder_NewPicture( p_dec );
    if( !p_pic )
    {
        msg_Err( p_dec, "Could not get picture" );
        goto error;
    }
    p_pic->b_progressive = p_sys->b_progressive;

    davinci_CopyXDMtoPicture( p_pic, &p_sys->out );

    /* */
    p_pic->date = p_block->i_pts > 0 ? p_block->i_pts : p_block->i_dts;

    if( p_block->i_buffer == 0 )
    {
        block_Release( p_block );
        *pp_block = NULL;
    }

    return p_pic;

error:
    if( p_pic )
        decoder_DeletePicture( p_dec, p_pic );

    block_Release( p_block );
    return NULL;
}
