/*****************************************************************************
 * davinci.c : DaVinci resizer module
 *****************************************************************************
 * Copyright (C) 2008-2010 M2X BV
 * $Id: 2d31dca2e53e5e6d594925d17089a0240ad84bcb $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *          Jean-Paul Saman <jean-paul saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#undef _FILE_OFFSET_BITS        /* mmap() will fail if we use 64bits offsets */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_filter.h>
#include <vlc_vout.h>

#include <string.h>                                            /* strerror() */
#include <stdlib.h>                                      /* malloc(), free() */
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/mman.h>

/* hmm, should have been a user headerfile */
#include <asm-arm/arch-davinci/davinci_resizer.h>

/*****************************************************************************
 * Local and extern prototypes.
 *****************************************************************************/
static int Open( filter_t * );
static void Close( filter_t * );

static void DAVINCI_RESIZER( filter_t *, picture_t *, picture_t * );
static picture_t *DAVINCI_RESIZER_Filter( filter_t *, picture_t * );

static uint8_t *GetBuffer( filter_t *, int buf_type, int size, int *offset );
static void FreeBuffer( filter_t *, uint8_t *p_buffer, int *i_size );

struct filter_sys_t
{
    int i_fd; /* /dev/davinci_resizer */

    rsz_resize_t resize;

    uint8_t *p_in;
    int i_in_size;
    uint8_t *p_out;
    int i_out_size;

    bool b_inited;
    bool b_scale;
};

/*****************************************************************************
 * Module descriptor.
 *****************************************************************************/
vlc_module_begin();
    set_description( N_("Conversions using /dev/davinci_resizer") )
    set_shortname( N_("Davinci Resizer" ) )
    set_capability( "video filter2", 200 )
    set_category( CAT_VIDEO )
    set_subcategory( SUBCAT_VIDEO_VFILTER )
    set_callbacks( Open, Close )
vlc_module_end();

/*****************************************************************************
 *
 *****************************************************************************/
static int Open( filter_t *p_this )
{
    filter_t *p_filter = (filter_t *)p_this;
    filter_sys_t *p_sys;

    switch( p_filter->fmt_in.video.i_chroma )
    {
#if 0
        case VLC_FOURCC('Y','V','1','2'):
        case VLC_FOURCC('I','4','2','0'):
        case VLC_FOURCC('I','Y','U','V'):
#endif
        case VLC_FOURCC('U','Y','V','Y'):
        case VLC_FOURCC('U','Y','N','V'):
        case VLC_FOURCC('Y','4','2','2'):
            switch( p_filter->fmt_out.video.i_chroma )
            {
                case VLC_FOURCC('U','Y','V','Y'):
                case VLC_FOURCC('U','Y','N','V'):
                case VLC_FOURCC('Y','4','2','2'):
                    p_filter->pf_video_filter = DAVINCI_RESIZER_Filter;
                    break;

                default:
                    return VLC_EGENERIC;
            }
            break;

        default:
            return VLC_EGENERIC;
    }

    p_filter->p_sys = p_sys = calloc( 1, sizeof( filter_sys_t ) );
    if( !p_sys )
        return VLC_ENOMEM;

    p_sys->i_fd = open( "/dev/davinci_resizer", O_RDWR );
    if( p_sys->i_fd == -1 )
    {
        msg_Err( p_filter, "Failed to open /dev/davinci_resizer (%m)." );
        free( p_sys );
        return VLC_EGENERIC;
    }

    p_sys->b_inited = false;
    p_sys->b_scale = config_GetInt( p_filter->p_libvlc, "fullscreen" );

    if( p_filter->fmt_in.video.i_chroma != p_filter->fmt_out.video.i_chroma )
        msg_Dbg( p_filter, "Davinci resizer %4.4s to %4.4s converter",
                 (const char*)&p_filter->fmt_in.video.i_chroma,
                 (const char*)&p_filter->fmt_out.video.i_chroma );
    else
        msg_Dbg( p_filter, "Davinci resizer %dx%d to %dx%d scaler",
                p_filter->fmt_in.video.i_width, p_filter->fmt_in.video.i_height,
                p_filter->fmt_out.video.i_width, p_filter->fmt_out.video.i_height );

    return VLC_SUCCESS;
}

static void Close( filter_t *p_filter )
{
    filter_sys_t *p_sys = p_filter->p_sys;

    FreeBuffer( p_filter, p_sys->p_in, &p_sys->i_in_size );
    FreeBuffer( p_filter, p_sys->p_out, &p_sys->i_out_size );
    close( p_sys->i_fd );
    free( p_sys );
}

/* Following functions are local */
static uint8_t *GetBuffer( filter_t *p_filter, int buf_type, int size,
                           int *offset )
{
    filter_sys_t *p_sys = p_filter->p_sys;
    uint8_t *p_buf;
    rsz_reqbufs_t reqbufs;
    rsz_buffer_t buffer;

    memset( &reqbufs, 0, sizeof( reqbufs ) );
    memset( &buffer, 0, sizeof( buffer ) );

    reqbufs.buf_type = buf_type;
    reqbufs.size = size;
    reqbufs.count = 1;

#if 0
    msg_Dbg( p_filter, "Asking %s for buffers: size 0x%x, count %d",
             buf_type == RSZ_BUF_IN ? "input" : "output",
             reqbufs.size, reqbufs.count );
#endif

    if( ioctl( p_sys->i_fd, RSZ_REQBUF, &reqbufs ) == -1 )
    {
        msg_Err( p_filter, "RSZ_REQBUF failed (%m)" );
        return NULL;
    }

    buffer.buf_type = buf_type;
    buffer.index = 0;

    if( ioctl( p_sys->i_fd, RSZ_QUERYBUF, &buffer ) == -1 )
    {
        msg_Err( p_filter, "RSZ_QUERYBUF failed (%m)" );
        return NULL;
    }

#if 0
    msg_Dbg( p_filter, "Got buffer: index %d, offset 0x%x",
             buffer.index, buffer.offset );
#endif

    if( offset ) *offset = buffer.offset;

    p_buf = mmap( NULL, reqbufs.size, PROT_READ|PROT_WRITE, MAP_SHARED,
                  p_sys->i_fd, buffer.offset );
    if( p_buf == MAP_FAILED )
    {
        msg_Err( p_filter, "mmap to buffer failed (%m)" );
        return NULL;
    }

    return p_buf;
}

static void FreeBuffer( filter_t *p_filter, uint8_t *p_buffer, int *i_size )
{
    if( p_buffer )
    {
        if( munmap( p_buffer, *i_size ) < 0 )
        {
            msg_Err( p_filter, "munmap on buffer failed (%m)" );
        }
        p_buffer = NULL;
        *i_size = 0;
    }
}

static int Init( filter_t *p_filter, picture_t *p_source, picture_t *p_dest )
{
    filter_sys_t *p_sys = p_filter->p_sys;
    int i;

    /**
     * Get the input and output buffers
     */
    if( !p_sys->p_in )
    {
        p_sys->i_in_size = p_source->p->i_pitch * p_source->p->i_visible_lines;
        p_sys->p_in = GetBuffer( p_filter, RSZ_BUF_IN, p_sys->i_in_size, &p_sys->resize.in_buf.offset );
        p_sys->resize.in_buf.index = 0;
        if( !p_sys->p_in )
            return VLC_EGENERIC;
    }

    if( !p_sys->p_out )
    {
        p_sys->i_out_size = p_dest->p->i_pitch * p_dest->p->i_visible_lines;
        p_sys->p_out = GetBuffer( p_filter, RSZ_BUF_OUT, p_sys->i_out_size, &p_sys->resize.out_buf.offset );
        p_sys->resize.out_buf.index = 0;
        if( !p_sys->p_out )
            return VLC_EGENERIC;
    }

    /**
     * Set the parameters
     */
    rsz_params_t params;
    memset( &params, 0, sizeof( params ) );
    params.in_hsize = p_source->p->i_visible_pitch >> 1;
    params.in_vsize = p_source->p->i_visible_lines;
    params.in_pitch = p_source->p->i_pitch;
    params.inptyp = RSZ_INTYPE_YCBCR422_16BIT;
    params.pix_fmt = RSZ_PIX_FMT_YUYV;
    params.out_hsize = p_dest->p->i_visible_pitch >> 1;
    params.out_vsize = p_dest->p->i_visible_lines;
    params.out_pitch = p_dest->p->i_pitch;

    const short coefs[] = {
        0, 256, 0, 0,
        -11, 245, 23, -1,
        -17, 220, 58, -5,
        -18, 184, 100, -10,
        -15, 143, 143, -15,
        -10, 100, 184, -18,
        -5, 58, 220, -17,
        -1, 23, 245, -11 };

    for( i = 0; i < 32; i++ )
        params.hfilt_coeffs[i] = params.vfilt_coeffs[i] = coefs[i];
    params.yenh_params.type = RSZ_YENH_DISABLE;

    if( ioctl( p_sys->i_fd, RSZ_S_PARAM, &params ) == -1 )
    {
        msg_Err( p_filter, "Failed setting resizer parameters (%m)" );
        return VLC_EGENERIC;
    }

    /**
     * Set resizer speed to fastest
     */
    int i_speed = 0;
    if( ioctl( p_sys->i_fd, RSZ_S_EXP, &i_speed ) == -1 )
    {
        msg_Err( p_filter, "Failed setting resizer speed (%m)" );
        return VLC_EGENERIC;
    }

    /* Initialize resize structure */
    p_sys->resize.in_buf.size = p_sys->i_in_size;
    p_sys->resize.out_buf.size = p_sys->i_out_size;

    p_filter->p_sys->b_inited = true;

    return VLC_SUCCESS;
}

/* Chroma conversion routine using Davinic Resizer */
VIDEO_FILTER_WRAPPER(DAVINCI_RESIZER)

static void DAVINCI_RESIZER( filter_t *p_filter, picture_t *p_source, picture_t *p_dest )
{
    filter_sys_t *p_sys = p_filter->p_sys;

    if( p_sys->b_scale ||
        p_filter->fmt_in.video.i_chroma != p_filter->fmt_out.video.i_chroma )
    {
        if( !p_sys->b_inited &&
            Init( p_filter, p_source, p_dest ) != VLC_SUCCESS )
            return;

        memcpy( p_sys->p_in, p_source->p->p_pixels, p_sys->i_in_size );

        if( ioctl( p_sys->i_fd, RSZ_RESIZE, &p_sys->resize ) )
        {
            msg_Err( p_filter, "Resize request failed." );
            return;
        }

        memcpy( p_dest->p->p_pixels, p_sys->p_out, p_sys->i_out_size );
    }

    if( !p_sys->b_scale &&
         p_filter->fmt_out.video.i_chroma == VLC_FOURCC('U','Y','V','Y') )
    {
        /* Do not resize the picture to fullscreen, but center it instead. */
        for( int i_plane = 0; i_plane < p_dest->i_planes; i_plane++ )
        {
            if( p_source->p[i_plane].i_pitch == p_dest->p[i_plane].i_pitch )
            {
                /* There are margins, but with the same width : perfect ! */
                vlc_memcpy( p_dest->p[i_plane].p_pixels, p_source->p[i_plane].p_pixels,
                            p_source->p[i_plane].i_pitch * p_source->p[i_plane].i_visible_lines );
            }
            else
            {
                uint8_t *p_in  = p_source->p[i_plane].p_pixels;
                uint8_t *p_out = p_dest->p[i_plane].p_pixels;

                int i_x_offset = (p_filter->fmt_out.video.i_width - p_source->format.i_width) /
                                 (p_source->p[i_plane].i_pixel_pitch);
                int i_y_offset = (p_filter->fmt_out.video.i_height - p_source->format.i_height) >> 1;

                int i_x = i_x_offset * p_source->p[i_plane].i_pixel_pitch;
                int i_x_clip, i_y_clip;

                /* Check boundaries, clip the image if necessary */
                i_x_clip = ( i_x + p_source->p[i_plane].i_visible_pitch ) - p_dest->p[i_plane].i_visible_pitch;
                i_x_clip = ( i_x_clip > 0 ) ? i_x_clip : 0;

                i_y_clip = ( i_y_offset + p_source->p[i_plane].i_visible_lines ) - p_dest->p[i_plane].i_visible_lines;
                i_y_clip = ( i_y_clip > 0 ) ? i_y_clip : 0;

                if( ( i_y_offset <= p_dest->p[i_plane].i_visible_lines ) &&
                    ( i_x <= p_dest->p[i_plane].i_visible_pitch ) )
                {
                    int i_line;

                    p_out += ( i_y_offset * p_dest->p[i_plane].i_pitch );
                    for( i_line = 0; i_line < ( p_source->p[i_plane].i_visible_lines - i_y_clip ); i_line++ )
                    {
                        vlc_memcpy( p_out + i_x, p_in,
                                    p_source->p[i_plane].i_visible_pitch - i_x_clip );
                        p_in  += p_source->p[i_plane].i_pitch;
                        p_out += p_dest->p[i_plane].i_pitch;
                    }
                }
            }
        }
    }
}
